import datetime
import unittest

from chia_tea.chia_watchdog.madmax.MadMaxPlotInProgress import (
    MadMaxPercentages,
    MadMaxPlotInProgress,
)
from .line_checks import (
    AddNewPlotInProgress,
    LatestPlotEnteringPhase2,
    LatestPlotEnteringPhase3,
    LatestPlotEnteringPhase4,
    SetFarmerPublicKeyForLatestPlot,
    SetLatestPlotProgressForPhase1,
    SetLatestPlotProgressForPhase2,
    SetLatestPlotProgressForPhase3,
    SetPlotDataForLatestPlot,
    SetPoolPublicKeyForLatestPlot,
    StartCopyOfPlot,
)
from ..ChiaWatchdog import ChiaWatchdog


class TestLineAction(unittest.TestCase):
    def setUp(self) -> None:
        self.public_key = "72fff9b4f41767c333314a126c06242297c3f20f44bcf1da1a82c7ec7eb8afaf"

    def test_adding_a_new_plot(self):

        logfile_line = "Process ID: 1374"

        dog = ChiaWatchdog("", "")

        action = AddNewPlotInProgress()
        self.assertTrue(action.is_match(logfile_line))
        some_time = datetime.datetime.now()
        action.apply(line=logfile_line, chia_dog=dog)
        self.assertEqual(len(dog.plots_in_progress), 1)
        plot = dog.plots_in_progress[0]
        self.assertEqual(plot.process_id, 1374)
        self.assertEqual(plot.state, "Init")
        self.assertTrue(plot.start_time > some_time)

    def test_plotting_plot_is_removed_after_finish(self):

        logfile_line = (
            f"Started copy to /some/drive/plot-k32-2021-10-06-00-53-{self.public_key}.plot"
        )

        dog = ChiaWatchdog("", "")
        dog.plots_in_progress.append(
            MadMaxPlotInProgress(
                process_id=1,
                public_key=self.public_key,
                farmer_public_key="",
                pool_public_key="",
                start_time=datetime.datetime.now(),
                progress=0.0,
                plot_type=32,
                state="",
            )
        )

        action = StartCopyOfPlot()
        self.assertTrue(action.is_match(logfile_line))
        action.apply(line=logfile_line, chia_dog=dog)
        self.assertEqual(dog.plots_in_progress, [])

    def test_pool_public_key_is_set_correctly(self):

        logfile_line = f"Pool Public Key:   {self.public_key}"

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=32,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = SetPoolPublicKeyForLatestPlot()
        self.assertTrue(action.is_match(logfile_line))
        action.apply(line=logfile_line, chia_dog=dog)
        self.assertEqual(dog.plots_in_progress, [plot])
        self.assertEqual(plot.pool_public_key, self.public_key)

    def test_farmer_public_key_is_set_correctly(self):

        logfile_line = f"Farmer Public Key: {self.public_key}"

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=32,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = SetFarmerPublicKeyForLatestPlot()
        self.assertTrue(action.is_match(logfile_line))
        action.apply(line=logfile_line, chia_dog=dog)
        self.assertEqual(dog.plots_in_progress, [plot])
        self.assertEqual(plot.farmer_public_key, self.public_key)

    def test_plot_data_is_set_correctly_for_latest_plot(self):

        logfile_line = f"Plot Name: plot-k32-2021-10-05-20-38-{self.public_key}"

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=0,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = SetPlotDataForLatestPlot()
        self.assertTrue(action.is_match(logfile_line))
        action.apply(line=logfile_line, chia_dog=dog)
        self.assertEqual(dog.plots_in_progress, [plot])
        self.assertEqual(plot.public_key, self.public_key)
        self.assertEqual(plot.start_time.year, 2021)
        self.assertEqual(plot.start_time.month, 10)
        self.assertEqual(plot.start_time.day, 5)
        self.assertEqual(plot.start_time.hour, 20)
        self.assertEqual(plot.start_time.minute, 38)
        self.assertEqual(plot.plot_type, 32)
        self.assertEqual(plot.state, "Plotting Phase1")

    def test_update_progress_for_phase1(self):

        lines = [
            "[P1] Table 1 took 56.4157 sec"
            "[P1] Table 2 took 312.581 sec, found 4294982064 matches"
            "[P1] Table 3 took 440.816 sec, found 4295062806 matches"
            "[P1] Table 4 took 443.39 sec, found 4294967684 matches"
            "[P1] Table 5 took 422.741 sec, found 4294973204 matches"
            "[P1] Table 6 took 368.958 sec, found 4294968504 matches"
            "[P1] Table 7 took 248.694 sec, found 4294916656 matches"
        ]

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=32,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = SetLatestPlotProgressForPhase1()
        for i_line, line in enumerate(lines):
            self.assertTrue(action.is_match(line))
            action.apply(line=line, chia_dog=dog)
            self.assertEqual(dog.plots_in_progress, [plot])
            self.assertEqual(plot.progress, MadMaxPercentages.phase1[i_line])

    def test_plot_entering_phase2(self):

        logfile_line = "[P2] max_table_size = 4295062806"

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=0,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = LatestPlotEnteringPhase2()
        self.assertTrue(action.is_match(logfile_line))
        action.apply(line=logfile_line, chia_dog=dog)
        self.assertEqual(dog.plots_in_progress, [plot])
        self.assertEqual(plot.state, "Plotting Phase2")

    def test_update_progress_for_phase2(self):

        lines = [
            "[P2] Table 7 scan took 39.6447 sec",
            "[P2] Table 7 rewrite took 134.675 sec, dropped 0 entries (0 %)",
            "[P2] Table 6 scan took 49.1251 sec",
            "[P2] Table 6 rewrite took 120.439 sec, dropped 581252134 entries (13.5333 %)",
            "[P2] Table 5 scan took 47.7513 sec",
            "[P2] Table 5 rewrite took 108.784 sec, dropped 761979853 entries (17.7412 %)",
            "[P2] Table 4 scan took 44.9049 sec",
            "[P2] Table 4 rewrite took 105.015 sec, dropped 828797543 entries (19.2969 %)",
            "[P2] Table 3 scan took 48.4281 sec",
            "[P2] Table 3 rewrite took 111.87 sec, dropped 855125503 entries (19.9095 %)",
            "[P2] Table 2 scan took 44.8467 sec",
            "[P2] Table 2 rewrite took 106.716 sec, dropped 865599957 entries (20.1538 %)",
        ]

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=32,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = SetLatestPlotProgressForPhase2()
        for i_line, line in enumerate(lines):
            self.assertTrue(action.is_match(line))
            action.apply(line=line, chia_dog=dog)
            self.assertEqual(dog.plots_in_progress, [plot])
            self.assertEqual(plot.progress, MadMaxPercentages.phase2[i_line])

    def test_plot_entering_phase3(self):

        logfile_line = "Wrote plot header with 268 bytes"

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=0,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = LatestPlotEnteringPhase3()
        self.assertTrue(action.is_match(logfile_line))
        action.apply(line=logfile_line, chia_dog=dog)
        self.assertEqual(dog.plots_in_progress, [plot])
        self.assertEqual(plot.state, "Plotting Phase3")

    def test_update_progress_for_phase3(self):

        lines = [
            "[P3-1] Table 2 took 119.469 sec, wrote 3429382107 right entries",
            "[P3-2] Table 2 took 114.577 sec, wrote 3429382107 left entries, 3429382107 final",
            "[P3-1] Table 3 took 143.732 sec, wrote 3439937303 right entries",
            "[P3-2] Table 3 took 131.013 sec, wrote 3439937303 left entries, 3439937303 final",
            "[P3-1] Table 4 took 132.11 sec, wrote 3466170141 right entries",
            "[P3-2] Table 4 took 158.274 sec, wrote 3466170141 left entries, 3466170141 final",
            "[P3-1] Table 5 took 133.552 sec, wrote 3532993351 right entries",
            "[P3-2] Table 5 took 140.348 sec, wrote 3532993351 left entries, 3532993351 final",
            "[P3-1] Table 6 took 141.1 sec, wrote 3713716370 right entries",
            "[P3-2] Table 6 took 145.006 sec, wrote 3713716370 left entries, 3713716370 final",
            "[P3-1] Table 7 took 177.476 sec, wrote 4294916656 right entries",
            "[P3-2] Table 7 took 162.498 sec, wrote 4294916656 left entries, 4294916656 final",
        ]

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=32,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = SetLatestPlotProgressForPhase3()
        for i_line, line in enumerate(lines):
            self.assertTrue(action.is_match(line))
            action.apply(line=line, chia_dog=dog)
            self.assertEqual(dog.plots_in_progress, [plot])
            self.assertEqual(plot.progress, MadMaxPercentages.phase3[i_line])

    def test_plot_entering_phase4(self):

        logfile_line = "[P4] Starting to write C1 and C3 tables"

        dog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            public_key=self.public_key,
            farmer_public_key="",
            pool_public_key="",
            start_time=datetime.datetime.now(),
            progress=0.0,
            plot_type=0,
            state="",
        )
        dog.plots_in_progress.append(plot)

        action = LatestPlotEnteringPhase4()
        self.assertTrue(action.is_match(logfile_line))
        action.apply(line=logfile_line, chia_dog=dog)
        self.assertEqual(dog.plots_in_progress, [plot])
        self.assertEqual(plot.state, "Plotting Phase4")



from typing import List

from chia_tea.discord.formatting import farmer_harvester_pb2_as_markdown

from ..common import catch_errors_as_message, open_database_read_only
from ..formatting import get_machine_info_name
from ..notifications.run_notifiers import get_current_computer_and_machine_infos_from_db


@catch_errors_as_message
async def farmers_cmd(db_filepath: str) -> List[str]:
    """Formats all farmers into a message

    Parameters
    ----------
    db_filepath : str
        path to the sqlite database

    Returns
    -------
    messages : List[str]
        list of messages to print
    """
    messages = []

    with open_database_read_only(db_filepath) as cursor:

        machine_and_computer_info_dict = get_current_computer_and_machine_infos_from_db(cursor)

        messages = []

        for _, (machine, computer_info) in machine_and_computer_info_dict.items():
            farmer_is_running = computer_info.farmer.is_running

            # a farmer is running, create a message
            if farmer_is_running:
                messages += [f"🧑‍🌾 *{get_machine_info_name(machine)}*:"]

                # list up connected harvesters
                messages += [
                    farmer_harvester_pb2_as_markdown(harvester)
                    for harvester in computer_info.farmer_harvesters
                ]

        if not messages:
            messages.append("No farmers 🧑‍🌾 around.")

    return messages



import ntpath
import os
import time

from ..utils.cli import parse_args
from ..utils.config import read_config
from ..utils.logger import get_logger
from .Disk import collect_files_from_folders, copy_file, find_disk_with_space
from .Lockfile import create_lockfile

module_name = "chia_tea.copy"


def main():
    """Main program for copying files"""

    # get command line arguments
    args = parse_args(
        name="Chia Tea Copy",
        description="Start a process copying files for your chia farm.",
    )

    # load config
    config = read_config(args.config)

    # get logger
    logger = get_logger(module_name)

    # little hack since "from" is a reserved keyword in python
    from_folders = config.copy.source_folders
    target_folders = config.copy.target_folders

    logger.info("Copying from: %s", from_folders)
    logger.info("Copying to  : %s", target_folders)

    # execute infinite copy loop
    while True:

        files_to_copy = collect_files_from_folders(from_folders, "*.plot")

        for filepath in files_to_copy:
            # search for a space on the specified disks
            target_dir = find_disk_with_space(target_folders, filepath)
            if target_dir is None:
                logger.error("No disk space available for: %s", filepath)
                continue

            # compose new filepath after move
            filename = ntpath.basename(filepath)
            target_path = os.path.join(target_dir, filename)

            # move file (with measurement)
            logger.info("moving file: %s -> %s", filepath, target_path)
            start = time.time()

            # lockfile
            path_to_lockfile = os.path.join(target_dir, filename.replace(".plot", ".copying"))

            with create_lockfile(path_to_lockfile):
                successful_copy = copy_file(filepath, target_path)

            duration_secs = time.time() - start
            if successful_copy:
                logger.info("done in %.1fs", duration_secs)
                try:
                    os.remove(filepath)
                except FileNotFoundError:
                    logger.error("Could not remove original file: %s", filepath)

            else:
                logger.error("failed to copy %s in %.1fs", filepath, duration_secs)

        # rate limiter
        time.sleep(15)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        get_logger(module_name).info("Shutting down.")



import logging
import os
import traceback

from ..utils.logger import get_logger


class Lockfile:
    """Class to manage lockfiles"""

    def __init__(self, path_to_lockfile: str):
        self.path = path_to_lockfile

    def __enter__(self):
        logging.info("Creating LogFile")
        with open(self.path, "w", encoding="utf8") as fp:
            try:
                fp.write("copying")
            except Exception:
                trace = traceback.format_stack()
                get_logger(__file__).error(trace)
                logging.error("Could not write to lockfile: %s", self.path)

    def __exit__(self, exception_type, exception_value, exception_traceback):
        logging.info("Removing LogFile")
        try:
            os.remove(self.path)
        except Exception:
            trace = exception_traceback.format_stack()
            get_logger(__file__).error(trace)
            logging.error("Could not remove lockfile: %s", self.path)


def create_lockfile(path_to_lockfile: str) -> Lockfile:
    """Creates a lockfile if used in a context through 'with'

    Parameters
    ----------
    path_to_lockfile : str
        filepath of lockfile

    Returns
    -------
    lockfile : Lockfile
        Lockfile instance which creates a lockfile if used in
        combination with a 'with' statement.
    """
    return Lockfile(path_to_lockfile)



from dataclasses import dataclass


@dataclass
class WalletAPI:
    """This class holds chia information fetched through RPC
    from chia services on the same machine
    """

    # pylint: disable=too-few-public-methods

    is_ready: bool = False
    n_wallets: int = 0
    is_synced: bool = False
    is_running: bool = False



from datetime import datetime
from typing import List

from ..protobuf.generated.chia_pb2 import (
    Harvester,
    HarvesterPlot,
    HarvesterViewedFromFarmer,
    PlotInProgress,
)
from ..protobuf.generated.hardware_pb2 import Cpu, Disk, Ram
from ..protobuf.generated.machine_info_pb2 import MachineInfo
from ..utils.timing import format_time_since, format_timedelta_from_secs


def format_memory_size(n_bytes: float, suffix: str = "B"):
    """Formats a memory size number

    Parameters
    ----------
    n_bytes : float
        bytes to format
    suffix : string
        suffix of the memory

    Notes
    -----
        Thanks Fred @ Stackoverflow:
        https://stackoverflow.com/questions/1094841/get-human-readable-version-of-file-size
    """
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(n_bytes) < 1024.0:
            return "%3.1f%s%s" % (n_bytes, unit, suffix)
        n_bytes /= 1024.0
    return "%.1f%s%s" % (n_bytes, "Yi", suffix)


def cpu_pb2_as_markdown(cpu: Cpu) -> str:
    """Formats a protobuf cpu as markdown

    Parameters
    ----------
    cpu : Cpu
        cpu instance with data

    Returns
    -------
    cpu_as_string : str
        cpu info as markdown string
    """

    return """   __CPU__
      name:  {name}
      cores: {cores}
      usage: {usage:.1f}%
      temp:  {temperature:.0f} deg
      clock: {speed:.0f} Mhz""".format(
        name=cpu.name,
        cores=cpu.n_vcores,
        usage=cpu.usage,
        temperature=cpu.temperature,
        speed=cpu.clock_speed,
    )


def ram_pb2_as_markdown(ram: Ram) -> str:
    """Formats a protobuf ram as markdown

    Parameters
    ----------
    ram : Ram
        ram instance with data

    Returns
    -------
    ram_as_string : str
        ram info as markdown string
    """

    used_percent = 0.0
    try:
        used_percent = ram.used_ram / ram.total_ram * 100
    except ZeroDivisionError:
        pass
    swap_percent = 0.0
    try:
        swap_percent = ram.used_swap / ram.total_swap * 100
    except ZeroDivisionError:
        pass

    return """   __RAM__
      total: {total}
      used:  {used:.1f}%
      swap:  {swap:.1f}%
      """.format(
        total=format_memory_size(ram.total_ram),
        used=used_percent,
        swap=swap_percent,
    )


def farmer_harvester_pb2_as_markdown(harvester: HarvesterViewedFromFarmer) -> str:
    """Formats a protobuf HarvesterViewedFromFarmer as markdown

    Parameters
    ----------
    harvester : HarvesterViewedFromFarmer
        farmer harvester instance with data

    Returns
    -------
    harvester_as_string : str
        farmer harvester info as markdown string
    """
    now_timestamp = datetime.now().timestamp()

    return """
  Harvester *{harvester_id}*
     🌐 ip address:  {ip_address}
     📡 last answer: {last_answer} ago
     🚆 missed challenges: {missed_challenges}
     🌾 plots: {n_plots}""".format(
        ip_address=harvester.ip_address,
        last_answer=format_timedelta_from_secs(now_timestamp - harvester.time_last_msg_received),
        harvester_id=harvester.id[:8],
        missed_challenges=harvester.missed_challenges,
        n_plots=harvester.n_plots,
    )


def disk_pb2_as_markdown(disk: Disk) -> str:
    """Formats a protobuf HarvesterViewedFromFarmer as markdown

    Parameters
    ----------
    disk : Disk
        disk instance with data

    Returns
    -------
    disk_as_string : str
        disk info as markdown string
    """
    usage_percent = 0
    try:
        usage_percent = disk.used_space / disk.total_space * 100
    except ZeroDivisionError:
        pass
    free_memory_as_str = format_memory_size(disk.total_space - disk.used_space)
    return f"        {usage_percent:3.1f}% {disk.id} ({free_memory_as_str} free)"


def harvester_pb2_as_markdown(
    machine: MachineInfo,
    harvester: Harvester,
    plots: List[HarvesterPlot],
    disks: List[Disk],
) -> str:
    """Formats a protobuf Harvester as markdown

    Parameters
    ----------
    machine : MachineInfo
        info about the machine including name and ip
    harvester : Harvester
        harvester instance with data
    plots : List[HarvesterPlot]
        plots of the harvester
    disks : List[Disk]
        disks of the harvester

    Returns
    -------
    harvester_as_string : str
        harvester info as markdown string
    """

    total_size = sum(plot.filesize for plot in plots)
    disk_msgs = [disk_pb2_as_markdown(disk) for disk in disks]

    return """
  🚜 Harvester {machine}
     🍀 proofs: {n_proofs}
     🌾 plots: {n_plots}
     🌾 size of plots: {total_size}
     💽 disks:
{disk_msgs}""".format(
        machine=get_machine_info_name(machine),
        n_plots=len(plots),
        n_proofs=harvester.n_proofs,
        total_size=format_memory_size(total_size),
        disk_msgs="\n".join(disk_msgs),
    )


def plot_in_progress_pb2_as_markdown(plot_in_progress: PlotInProgress) -> str:
    """Formats a protobuf PlotInProgress as markdown

    Parameters
    ----------
    plot_in_progress: PlotInProgress
        plot in progress to be formatted in markdown

    Returns
    -------
    msg : str
        plot_in_progress as markdown string
    """

    start_time_dt = datetime.fromtimestamp(plot_in_progress.start_time)

    return "\n".join(
        (
            f"  - 🌽 Plot {plot_in_progress.id[:12]}...",
            f"       Since: {format_time_since(start_time_dt)}",
            f"       State: {plot_in_progress.state}",
            f"       Progress: {plot_in_progress.progress*100:.1f}%",
        )
    )


def get_machine_info_name(machine: MachineInfo) -> str:
    """Get a nicely formatted name for a machine info

    Parameters
    ----------
    machine : MachineInfo
        machine info to get a name of

    Returns
    -------
    name : str
        nicely formatted name of the machine info
    """

    return "{name} {id} ({ip})".format(
        name=f"{machine.name} -" if machine.name else "",
        id=str(machine.machine_id)[:10],
        ip=machine.ip_address,
    )



import glob
import os
import random
import shutil
import traceback
from typing import Dict, List, Union

import psutil

from ..utils.logger import get_logger


def filter_least_used_disks(disk_to_lockfile_count: Dict[str, int]) -> List[str]:
    """Filters for the least used disk

    Parameters
    ----------
    disk_to_lockfile_count : Dict[str, int]
        Dictionary of disks with lockfile count

    Returns
    -------
    available_dirs : List[str]
        Available directories with minimal lockfile count. In case of
        equal lockfile count then the disks are given randomly.
    """
    minimum_number_of_lockfiles = min(disk_to_lockfile_count.values())
    available_target_dirpaths = [
        dirpath
        for dirpath in disk_to_lockfile_count
        if disk_to_lockfile_count[dirpath] == minimum_number_of_lockfiles
    ]
    random.shuffle(available_target_dirpaths)
    return available_target_dirpaths


def find_disk_with_space(target_dirs: List[str], filepath_file: str) -> Union[str, None]:
    """Searches for space for a file to be moved

    Parameters
    ----------
    target_dirs: List[str]
        Directories in which the file can be copied
    filepath_file: str
        Path to the file to be copied

    Returns
    -------
    dirpath: Union[str, None]
        A target dir with space or None if no space available
    """
    logger = get_logger(__file__)
    fstat = os.stat(filepath_file)
    disk_to_lockfile_count = update_lockfile_count(target_dirs)
    available_target_dirpaths = filter_least_used_disks(disk_to_lockfile_count)

    for dirpath in available_target_dirpaths:
        try:
            if not os.path.isdir(dirpath):
                os.makedirs(dirpath, exist_ok=True)
            # size check
            n_lockfiles = disk_to_lockfile_count[dirpath]
            space_after_copying = n_lockfiles * 1.08e11  # 108GB
            space = psutil.disk_usage(dirpath)
            if space.free > (fstat.st_size + space_after_copying):
                return dirpath
        except PermissionError:
            warn_msg = "Permission denied to directory '%s'."
            logger.warning(warn_msg, dirpath)
        except FileNotFoundError:
            warn_msg = "Directory '%s' does not exist."
            logger.warning(warn_msg, dirpath)
        except ConnectionResetError:
            warn_msg = "Lost connection to network drive '%s'"
            logger.warning(warn_msg, dirpath)
        except OSError:
            warn_msg = "Cannot reach host for drive '%s'"
            logger.warning(warn_msg, dirpath)

    return None


def copy_file(source_path: str, target_path: str) -> bool:
    """Copies a file from a source path to a target path

    Parameters
    ----------
    source_path : str
        Path to the existing source file
    target_path: str
        Path where to copy the file

    Returns
    -------
    success : bool
        If the copy was a success.
    """
    print(target_path)
    with open(source_path, "rb") as fin:
        with open(target_path, "wb") as fout:
            try:
                shutil.copyfileobj(fin, fout, 128 * 1024)
            except (Exception, ConnectionResetError):
                trace = traceback.format_stack()
                get_logger(__file__).error(trace)
                return False
    return True


def collect_files_from_folders(folder_list: List[str], pattern: str) -> List[str]:
    """Collect files from folders

    Parameters
    ----------
    folder_list : List[str]
        list of folders to search for files
    pattern : str
        file pattern to search for

    Returns
    -------
    filepaths : List[str]
        paths to the files
    """
    logger = get_logger(__file__)

    all_filepaths = []

    for folder_path in folder_list:

        if not os.path.isdir(folder_path):
            if os.path.isfile(folder_path):
                warn_msg = "Path '{0}' is a file and not a directory."
            else:
                warn_msg = "Folder '{0}' does not exist or is not a directory."
            logger.warning(warn_msg, folder_path)
            continue

        all_filepaths += glob.glob(os.path.join(folder_path, pattern))

    return all_filepaths


def update_lockfile_count(target_dirs: List[str]) -> Dict[str, int]:
    """Get the lockfile count for the specified directories

    Parameters
    ----------
    target_dirs : List[str]
        Directories to get lockfile count for.

    Returns
    -------
    lockfile_count : Dict[str, int]
        Dictionary containing as key the directory and as
        value den lockfile count.
    """
    n_lockfiles_per_disk = {}
    for target_dir in target_dirs:
        lock_files = collect_files_from_folders([target_dir], "*.copying")
        n_lockfiles = len(lock_files)
        n_lockfiles_per_disk[target_dir] = n_lockfiles
    return n_lockfiles_per_disk



import unittest

from google.protobuf.json_format import ParseDict

from ...protobuf.generated.chia_pb2 import HarvesterViewedFromFarmer
from ...protobuf.generated.computer_info_pb2 import ComputerInfo
from ...protobuf.generated.machine_info_pb2 import MachineInfo
from .computer_info_notifications import (
    HARVESTER_TIMOUT,
    get_msg_if_farmer_harvester_timed_out,
    notify_on_harvester_reward_found,
    notify_on_wallet_connection_change,
    notify_on_wallet_sync_change,
)


class TestComputerInfoNotifications(unittest.TestCase):
    def test_farmer_harvester_reward_found_notification(self):

        machine = MachineInfo()

        old_computer_info = ParseDict(
            js_dict={"harvester": {"n_proofs": 0}},
            message=ComputerInfo(),
        )
        new_computer_info = ParseDict(
            js_dict={"harvester": {"n_proofs": 1}},
            message=ComputerInfo(),
        )

        messages = notify_on_harvester_reward_found(
            machine,
            old_computer_info=old_computer_info,
            new_computer_info=new_computer_info,
        )

        self.assertEqual(len(messages), 1)

    def test_farmer_harvester_newly_registered_and_reward_found(self):

        machine = MachineInfo()

        old_computer_info = ParseDict(
            js_dict={},
            message=ComputerInfo(),
        )
        new_computer_info = ParseDict(
            js_dict={"harvester": {"n_proofs": 1}},
            message=ComputerInfo(),
        )

        messages = notify_on_harvester_reward_found(
            machine,
            old_computer_info=old_computer_info,
            new_computer_info=new_computer_info,
        )

        self.assertEqual(len(messages), 1)

    def test_farmer_harvester_no_reward_found_notification(self):

        machine = MachineInfo()

        old_computer_info = ParseDict(
            js_dict={"harvester": {"n_proofs": 0}},
            message=ComputerInfo(),
        )
        new_computer_info = ParseDict(
            js_dict={"harvester": {"n_proofs": 0}},
            message=ComputerInfo(),
        )

        messages = notify_on_harvester_reward_found(
            machine,
            old_computer_info=old_computer_info,
            new_computer_info=new_computer_info,
        )

        self.assertEqual(len(messages), 0)

    def test_msg_if_wallet_not_synced_anymore(self):

        machine = MachineInfo()

        old_computer_info = ParseDict(
            js_dict={"wallet": {"is_synced": True}},
            message=ComputerInfo(),
        )
        new_computer_info = ParseDict(
            js_dict={"wallet": {"is_synced": False}},
            message=ComputerInfo(),
        )

        messages = notify_on_wallet_sync_change(
            machine=machine,
            old_computer_info=old_computer_info,
            new_computer_info=new_computer_info,
        )

        self.assertEqual(len(messages), 1)

    def test_msg_if_wallet_not_synced_again(self):

        machine = MachineInfo()

        old_computer_info = ParseDict(
            js_dict={"wallet": {"is_synced": False}},
            message=ComputerInfo(),
        )
        new_computer_info = ParseDict(
            js_dict={"wallet": {"is_synced": True}},
            message=ComputerInfo(),
        )

        messages = notify_on_wallet_sync_change(
            machine=machine,
            old_computer_info=old_computer_info,
            new_computer_info=new_computer_info,
        )

        self.assertEqual(len(messages), 1)

    def test_msg_if_wallet_has_started(self):

        machine = MachineInfo()

        old_computer_info = ParseDict(
            js_dict={"wallet": {"is_running": False}},
            message=ComputerInfo(),
        )
        new_computer_info = ParseDict(
            js_dict={"wallet": {"is_running": True}},
            message=ComputerInfo(),
        )

        messages = notify_on_wallet_connection_change(
            machine=machine,
            old_computer_info=old_computer_info,
            new_computer_info=new_computer_info,
        )

        self.assertEqual(len(messages), 1)

    def test_msg_if_wallet_has_stopped(self):

        machine = MachineInfo()

        old_computer_info = ParseDict(
            js_dict={"wallet": {"is_running": True}},
            message=ComputerInfo(),
        )
        new_computer_info = ParseDict(
            js_dict={"wallet": {"is_running": False}},
            message=ComputerInfo(),
        )

        messages = notify_on_wallet_connection_change(
            machine=machine,
            old_computer_info=old_computer_info,
            new_computer_info=new_computer_info,
        )

        self.assertEqual(len(messages), 1)

    def test_notification_on_timeout(self):

        machine = MachineInfo()

        # we dont perform a unittest on the notifier itself
        # since it's timing logic is hard to test. Instead
        # we test the internal wrapper function.

        # no msg case if all is good
        barely_ok_farmer_harvester = ParseDict(
            js_dict={"time_last_msg_received": 1},
            message=HarvesterViewedFromFarmer(),
        )
        msg = get_msg_if_farmer_harvester_timed_out(
            harvester=barely_ok_farmer_harvester,
            last_timestamp=59,
            new_timestamp=60,
            machine=machine,
        )
        self.assertTrue(msg == "")

        # on startup no message is given if a harvester
        # is already timed out to avoid spamming the chat
        # on restarts of the bot.
        msg_timestamp = 1
        timed_out_farmer_harvester = ParseDict(
            js_dict={"time_last_msg_received": msg_timestamp},
            message=HarvesterViewedFromFarmer(),
        )
        msg = get_msg_if_farmer_harvester_timed_out(
            harvester=timed_out_farmer_harvester,
            last_timestamp=0,
            new_timestamp=msg_timestamp + HARVESTER_TIMOUT + 2,
            machine=machine,
        )
        self.assertTrue(msg == "")

        # test general timeout notification mechanism
        msg = get_msg_if_farmer_harvester_timed_out(
            harvester=timed_out_farmer_harvester,
            last_timestamp=msg_timestamp + HARVESTER_TIMOUT - 1,
            new_timestamp=msg_timestamp + HARVESTER_TIMOUT,
            machine=machine,
        )
        self.assertTrue(len(msg) != 0)



from typing import Dict, Any
import os
import json

settings_dir = os.path.expanduser("~/.chia_tea")
path_to_settings = os.path.join(settings_dir, "settings.json")

# for caching settings locally
SETTINGS_CACHE = {}
SETTINGS_ARE_LOADED = False


def ensure_chia_settings_file():
    """Ensure the chia settings file exists"""
    if not os.path.isdir(settings_dir):
        os.makedirs(settings_dir, exist_ok=True)

    if not os.path.isfile(path_to_settings):
        with open(path_to_settings, "w", encoding="utf8") as file_handle:
            file_handle.write(json.dumps(SETTINGS_CACHE))


def read_settings_file() -> Dict[str, str]:
    """Reads the settings File and puts its content to the settings variable"""
    ensure_chia_settings_file()

    with open(path_to_settings, "r", encoding="utf8") as file_handle:
        return json.loads(file_handle.read())


def write_settings_file():
    """Writes the settings File to disc"""
    with open(path_to_settings, "w", encoding="utf8") as settings_file:
        json.dump(SETTINGS_CACHE, settings_file)


def ensure_settings_are_loaded():
    """Checks that the settings are loaded"""
    if not SETTINGS_ARE_LOADED:
        # pylint: disable=global-statement
        global SETTINGS_CACHE
        SETTINGS_CACHE = read_settings_file()


def get_settings_value(key: str, default: Any) -> Any:
    """Returns the value of a key in the settings Dict

    Parameters
    ----------
    key : str
        key of which the value is returned of
    default : str
        default value to get

    Returns
    -------
    value : Any
        Either value from settings or default value
    """
    ensure_settings_are_loaded()

    value = SETTINGS_CACHE.get(key)

    if value is None:
        set_settings_value(key, default)
        value = default

    return value


def set_settings_value(name: str, value: Any):
    """Sets the value of the Dict and writes settings to disc

    Parameters
    ----------
    name : str
        name of the attribute to set
    value : Any
        value to set
    """
    ensure_settings_are_loaded()
    SETTINGS_CACHE[name] = value
    write_settings_file()



from typing import Dict
import psutil

from .ChiaWatchdog import ChiaWatchdog


# pylint: disable=unused-argument
def remove_plotting_plots_if_madmax_does_not_run(watchdog: ChiaWatchdog):
    """Removes the plotting plots if the plotting process was killed

    Parameters
    ----------
    watchdog : ChiaWatchdog
        Watchdog to run regular checks on
    """
    process_ids = [plot.process_id for plot in watchdog.plots_in_progress if plot.process_id > 0]

    process_cache: Dict[int, psutil.Process] = {}
    dead_process_ids = []
    for pid in process_ids:
        try:
            process = process_cache.get(pid) or psutil.Process(pid)
            process_cache[process.pid] = process
            if process.is_running() and "chia_plot" in process.exe():
                # all good
                pass
            else:
                dead_process_ids.append(pid)
        except psutil.NoSuchProcess:
            dead_process_ids.append(pid)

    watchdog.plots_in_progress = [
        plot for plot in watchdog.plots_in_progress if plot.process_id not in dead_process_ids
    ]


async def run_watchdog_checks(watchdog: ChiaWatchdog):
    """Runs the harvester checks. Checks allow modification.
    Parameters
    ----------
    watchdog : ChiaWatchdog
        Watchdog to run regular checks on
    """

    for check_func in ALL_HARVESTER_CHECKS:
        check_func(watchdog=watchdog)


ALL_HARVESTER_CHECKS = [remove_plotting_plots_if_madmax_does_not_run]



import asyncio

from ..ChiaWatchdog import ChiaWatchdog
from .update_from_farmer import update_from_farmer
from .update_from_harvester import update_from_harvester
from .update_from_wallet import update_from_wallet


async def update_directly_from_chia(chia_dog: ChiaWatchdog):
    """Update the chia watchdog directly with data received from chia

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        watchdog instance to be modified
    """
    await asyncio.gather(
        update_from_farmer(chia_dog=chia_dog),
        update_from_wallet(chia_dog=chia_dog),
        update_from_harvester(chia_dog=chia_dog),
    )



import asyncio
from datetime import datetime, timedelta
from functools import wraps
from typing import Optional, Union

MILLISECOND = 1e-3
MICROSECONDS = 1e-6
SECOND = 1
MINUTE = 60
HOUR = 60 * 60
DAY = 24 * 60 * 60


def format_timedelta_from_secs(seconds: Union[float, None]) -> str:
    """Format datetime into a better format

    Parameters
    ----------
    seconds : Union[float, None]
        timedelta in seconds

    Returns
    -------
    response: str
        nicely formatted timedelta
    """
    # pylint: disable=too-many-return-statements

    if seconds is None:
        return ""

    if seconds < 0:
        return ""

    response_short = "{0:.0f}{1}"
    response_long = "{0:.2f}{1}"

    if seconds < MILLISECOND:
        return response_short.format(seconds / MICROSECONDS, "µs")

    if seconds < SECOND:
        return response_short.format(seconds / MILLISECOND, "ms")

    if seconds < MINUTE:
        return response_long.format(seconds / SECOND, "s")

    if seconds < HOUR:
        return response_long.format(seconds / MINUTE, "m")

    # if seconds < DAY:
    return response_long.format(seconds / HOUR, "h")


def format_timedelta(delta_time: Union[timedelta, None]) -> str:
    """Format datetime into a better format

    Parameters
    ----------
    delta_time: Union[timedelta, None]
        timedelta object to format

    Returns
    -------
    response: str
        datetime object nicely formatted
    """

    if delta_time is None:
        return ""

    dt_secs = delta_time.total_seconds()

    return format_timedelta_from_secs(dt_secs)


def format_time_since(start_time: Optional[datetime]) -> str:
    """Format the time since start nicely

    Parameters
    ----------
    start_time : Optional[datetime]
        time the counting started

    Returns
    -------
    reponse : str
        time duration as string

    Notes
    -----
        Returns empty string on negative time.
    """

    if start_time is None:
        return ""

    return format_timedelta(datetime.now() - start_time)


async def async_throttle(min_duration: float):
    """Logs the time of the function call

    Parameters
    ----------
    min_duration : float
        minimum duration the function will take
    """

    def decorator(function):
        @wraps(function)
        async def _throttle_it(*args, **kwargs):
            start_time = datetime.now()
            try:
                return await function(*args, **kwargs)
            finally:
                end_time = datetime.now()
                duration = (end_time - start_time).total_seconds()
                wait_duration = max(0, min_duration - duration)
                await asyncio.sleep(wait_duration)

        return _throttle_it

    return decorator


async def wait_at_least(min_duration: float, start_time: Optional[datetime] = None):
    """This function delays until at least the duration has
    passed since the specified start time

    Parameters
    ----------
    min_duration : float
        minimum duration to wait since the start time
    start_time : Optional[datetime]
        optional starting time. Will be set as now if not specified.
    """

    start_time = datetime.now() if start_time is None else start_time
    duration = (datetime.now() - start_time).total_seconds()

    await asyncio.sleep(max(0, min_duration - duration))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chia_tea',
 'chia_tea.chia_watchdog',
 'chia_tea.chia_watchdog.api',
 'chia_tea.chia_watchdog.logfile',
 'chia_tea.chia_watchdog.madmax',
 'chia_tea.config',
 'chia_tea.copy_plots',
 'chia_tea.discord',
 'chia_tea.discord.commands',
 'chia_tea.discord.notifications',
 'chia_tea.monitoring',
 'chia_tea.protobuf',
 'chia_tea.protobuf.data_collection',
 'chia_tea.protobuf.generated',
 'chia_tea.protobuf.to_sqlite',
 'chia_tea.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'aiohttp==3.7.4',
 'chia-blockchain>=1.2.3,<2.0.0',
 'concurrent-log-handler>=0.9.19,<0.10.0',
 'discord>=1.7.3,<2.0.0',
 'grpcio>=1.39.0,<2.0.0',
 'psutil>=5.8.0,<6.0.0',
 'py-cpuinfo>=8.0.0,<9.0.0',
 'rich>=10.10.0,<11.0.0',
 'sortedcontainers==2.3.0']

extras_require = \
{':sys_platform == "win32"': ['pywin32==300']}

setup_kwargs = {
    'name': 'chia-tea',
    'version': '0.2.0rc8',
    'description': 'A library dedicated to chia-blockchain farmer.',
    'long_description': '# Chia Tea 🌱🍵\n\nChia Tea is a tools and utility library for the Chia Cryptocurrency.\nWe are building tools in this library to serve our own needs and share our work openly with others.\nFeel free to use them and make your life easier 💚\n\n# Table of Contents\n\n- [Intro and FAQ](docs/faq.md)\n- [Quick-Start](docs/quick_start.md)\n- [Architecture Summary](docs/architecture_summary.md)\n- [Structure of this repository](#structure-of-this-repository)\n- [Feature Overview](#feature-overview)\n  - [Copy Tool](#copy-tool)\n  - [Monitoring](#monitoring)\n  - [Discord Bot](#discord-bot)\n- [Contributions](#contributions)\n- [About Us](#about-us)\n- [✊ Support Us ✊](#support-us)\n\n## Structure of this repository\n\n![Chia-Tea Infrastructure](docs/Chia_Infrastructure.png?raw=true)\n\n## Feature Overview\n\n### Copy Tool\n\nCopy is a tool to copy your chia files to a different location. It can be faster to plot to a temporary storage space and then move the plots to your harvester afterwards to not block the plotting queue. We manage this process through our copy cli tool. It incorporates the following features:\n\n- Selects a drive with sufficient space from multiple disks specified\n- Checks drive space regularly\n- Takes plots which are being copied already into account\n- Uses the drive with the fewest copy processes\n- Logs transfer times\n\n### Monitoring\n\nThe monitoring tracks everything relevant to a Chia farm, including harvesters, farmers, plotters, etc. It consists of a server and multiple clients. The server is run on only one machine. It stores all the monitoring data in a sqlite database. The clients are the data collectors and are typically run on plotters, harvesters, farmers, etc. Simply start them on the same machine and they will collect data automatically. You can run both server and client on the same machine if you have a single machine setup.\n\n### Discord Bot\n\nThe discord module is a bot watching a farm and reports major incidents. Internally the bot keeps an eye on the sqlite database thus a running monitoring server is required. Besides notifications the bot also provides commands to interact with the chia farm.\n\n<a name="about-us"></a>\n\n## About Us\n\nWe are a small group of professional engineers and software developers doing Chia for fun. Join our [Discord Server](https://discord.gg/kUS8AQEzsC) for questions, tips and tricks or just come over for a nice warm cup of your favourite tea.\n\n<a name="security"></a>\n\n## Contributions\n\nDue to security concerns, we only accept small PR\'s with limited complexity to our codebase.\n\n<a name="support-us"></a>\n\n## ✊ Support Us ✊\n\nThe more you support us, the easier we can make your life as a Chia farmer. Every little bit helps and motivates us to do more.\n\n| Currency | Address                                                                                                 |\n| -------- | ------------------------------------------------------------------------------------------------------- |\n| Chia     | xch13yrhjp0zleepsafjh8syh0jyakjgat9fzlut575lq0z5jywmydeqy05awj                                          |\n| BTC      | bc1qwjyh0fu708zv0yqdmp098tq465qy64jpmqpj4y                                                              |\n| ETH      | 0xeeaA95F8816208b4bb8D070ab571941843246029                                                              |\n| ADA      | addr1qxc2amr663yfh9z4cdk8d6hkv9apvm35dm5lkgjdlu6ffkfggrvustlynuxzqmswee4mvd6cfeu66hq788rmgts2uggq7qtuqh |\n',
    'author': 'Tea n Tech',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.youtube.com/channel/UCba194Pls_bHSqWoWMGoyzA',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os
import traceback
from datetime import datetime
from typing import Tuple

from ...utils.logger import get_logger
from ..ChiaWatchdog import ChiaWatchdog
from ..logfile.line_checks import AbstractLineAction
from .MadMaxPlotInProgress import MadMaxPercentages, MadMaxPlotInProgress


class AddNewPlotInProgress(AbstractLineAction):
    """Add a new plot in progress

    Example Line:
    Process ID: 1374

    Note
    ----
        We could add a new plot on the log cmd before
        but we don't have a process id then and if the
        plotting process dies we can't verify this and
        have zombie plots in progress.
    """

    LINE_START = "Process ID:"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):

        # Formatter conflict with space before :
        line = line[len(self.LINE_START) :].strip()  # noqa: E203
        process_id = int(line)

        chia_dog.plots_in_progress.append(
            MadMaxPlotInProgress(
                process_id=process_id,
                public_key="",
                pool_public_key="",
                farmer_public_key="",
                start_time=datetime.now(),
                progress=0,
                plot_type=0,
                state="Init",
            )
        )


class SetPoolPublicKeyForLatestPlot(AbstractLineAction):
    """Set the public key for the latest plot in progress

    Example Line:
    Pool Public Key:   ...
    """

    LINE_START = "Pool Public Key:"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        # Formatter conflict with space before :
        line = line[len(self.LINE_START) :].strip()  # noqa: E203

        if chia_dog.plots_in_progress:
            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.pool_public_key = line


class SetFarmerPublicKeyForLatestPlot(AbstractLineAction):
    """Set the farmer public key for the latest plot

    Example Line:
    Farmer Public Key: ...
    """

    LINE_START = "Farmer Public Key:"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        # Formatter conflict with space before :
        line = line[len(self.LINE_START) :].strip()  # noqa: E203

        if chia_dog.plots_in_progress:
            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.farmer_public_key = line


class SetPlotDataForLatestPlot(AbstractLineAction):
    """Extracts plot data from plot name

    Example Line:
    Plot Name: plot-k32-2021-10-05-22-02-{public key}
    """

    LINE_START = "Plot Name:"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        # Formatter conflict with space before :
        line = line[len(self.LINE_START) :].strip()  # noqa: E203

        if chia_dog.plots_in_progress:
            plot_name_split = line.split("-")

            _, plot_type, year, month, day, hour, minute, public_key = plot_name_split

            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.public_key = public_key
            latest_plot.start_time = datetime(
                int(year), int(month), int(day), int(hour), int(minute)
            )
            # remove the k before e.g. k32
            latest_plot.plot_type = int(plot_type[1:])
            latest_plot.state = "Plotting Phase1"


class SetLatestPlotProgressForPhase1(AbstractLineAction):
    """Set the progress for updates during phase 1

    Example Lines:
    [P1] Table 1 took 149.086 sec
    [P1] Table 2 took 313.995 sec, found 4294938576 matches
    [P1] Table 3 took 453.323 sec, found 4294895885 matches
    [P1] Table 4 took 435.455 sec, found 4294851544 matches
    [P1] Table 5 took 392.077 sec, found 4294722099 matches
    [P1] Table 6 took 352.225 sec, found 4294381733 matches
    [P1] Table 7 took 260.078 sec, found 4293784356 matches
    """

    LINE_START = "[P1] Table"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        if chia_dog.plots_in_progress:
            line_split = line.split()
            table_index = int(line_split[2]) - 1

            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.progress = MadMaxPercentages.phase1[table_index]


class LatestPlotEnteringPhase2(AbstractLineAction):
    """Set the progress for updates during phase 2

    Example Line:
    [P2] max_table_size = 4294967296
    """

    LINE_START = "[P2] max_table_size"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        if chia_dog.plots_in_progress:
            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.state = "Plotting Phase2"


class SetLatestPlotProgressForPhase2(AbstractLineAction):
    """Sets the progress for phase 2

    Steps are:
    [P2] Table 7 scan ...
    [P2] Table 7 rewrite ...
    [P2] Table 6 scan ...
    [P2] Table 6 rewrite ...
    [P2] Table 5 scan ...
    [P2] Table 5 rewrite ...
    [P2] Table 4 scan ...
    [P2] Table 4 rewrite ...
    [P2] Table 3 scan ...
    [P2] Table 3 rewrite ...
    [P2] Table 2 scan ...
    [P2] Table 2 rewrite ...
    """

    LINE_START = "[P2] Table"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        if chia_dog.plots_in_progress:
            line_split = line.split()
            table_index = int(line_split[2]) - 1
            one_more = line_split[3] == "rewrite"
            phase2_index = 11 - 2 * table_index + one_more + 1

            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.progress = MadMaxPercentages.phase2[phase2_index]


class LatestPlotEnteringPhase3(AbstractLineAction):
    """Update that the latest plot is entering phase 3

    Example Line:
    Wrote plot header with 268 bytes
    """

    LINE_START = "Wrote plot header with"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        if chia_dog.plots_in_progress:
            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.state = "Plotting Phase3"


class SetLatestPlotProgressForPhase3(AbstractLineAction):
    """Sets the progress for phase 2

    Steps are:
    [P3-1] Table 2 ...
    [P3-2] Table 2 ...
    [P3-1] Table 3 ...
    [P3-2] Table 3 ...
    [P3-1] Table 4 ...
    [P3-2] Table 4 ...
    [P3-1] Table 5 ...
    [P3-2] Table 5 ...
    [P3-1] Table 6 ...
    [P3-2] Table 6 ...
    [P3-1] Table 7 ...
    [P3-2] Table 7 ...
    """

    def is_match(self, line: str) -> bool:
        return line.startswith("[P3-1]") or line.startswith("[P3-2]")

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        if chia_dog.plots_in_progress:
            line_split = line.split()
            offset = 2
            table_index = int(line_split[2]) - offset
            one_more = line_split[0][4] == "2"
            phase3_index = 2 * table_index + one_more

            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.progress = MadMaxPercentages.phase3[phase3_index]


class LatestPlotEnteringPhase4(AbstractLineAction):
    """Set latest plot entering phase 4

    Example Line:
    [P4] Starting to write C1 and C3 tables
    """

    LINE_START = "[P4] Starting to write"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        if chia_dog.plots_in_progress:
            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.state = "Plotting Phase4"


class SetLatestPlotAsFinished(AbstractLineAction):
    """Mark the latest plot as finished

    Example Line:
    Total plot creation time was 5129.61 sec (85.4936 min)
    """

    LINE_START = "Total plot creation time was"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        if chia_dog.plots_in_progress:
            latest_plot = chia_dog.plots_in_progress[-1]
            latest_plot.state = "Plotting Done"
            latest_plot.progress = 1


class StartCopyOfPlot(AbstractLineAction):
    """
    Example line:
    Started copy to /some/drive/plot-k32-2021-10-06-00-53-{public_key}.plot
    """

    LINE_START = "Started copy to"

    def is_match(self, line: str) -> bool:
        return line.startswith(self.LINE_START)

    def apply(self, line: str, chia_dog: ChiaWatchdog):
        line_split = line.split()
        # get plot filepath from line
        plot_filepath = line_split[3]
        # isolate the plot name
        plot_name = os.path.basename(plot_filepath)
        # cut out the public key
        public_key_with_file_ending = plot_name.split("-")[7]
        # remove file ending from public key string
        public_key, _ = os.path.splitext(public_key_with_file_ending)

        # remove plot from plotting list
        chia_dog.plots_in_progress = [
            plot for plot in chia_dog.plots_in_progress if plot.public_key != public_key
        ]


# class FinishedCopyOfPlot(AbstractLineAction):
#     LINE_START = "Copy to"

#     def is_match(self, line: str) -> bool:
#         return line.startswith(self.LINE_START)

#     def apply(self, line: str, chia_dog: ChiaWatchdog):
#         line_split = line.split()
#         plot_filepath = line_split[2]
#         plot_name = os.path.basename(plot_filepath)
#         public_key = plot_name.split("-")[7]

#         for plot in chia_dog.plots_in_progress:
#             if plot.public_key == public_key:
#                 plot.end_time_copy = datetime.now()
#                 break


async def run_line_checks(chia_dog: ChiaWatchdog, line: str):
    """Processes a line from the logfile

    Parameters
    ----------
    line : str
        logfile line
    """
    try:
        if line:
            for action in ALL_LINE_ACTIONS:
                if action.is_match(line):
                    action.apply(line, chia_dog)

    except Exception:
        trace = traceback.format_exc()
        err_msg = "Error in line: %s\n%s"
        get_logger(__name__).error(err_msg, line, trace)


ALL_LINE_ACTIONS: Tuple[AbstractLineAction, ...] = (
    AddNewPlotInProgress(),
    SetPoolPublicKeyForLatestPlot(),
    SetFarmerPublicKeyForLatestPlot(),
    SetPlotDataForLatestPlot(),
    SetLatestPlotProgressForPhase1(),
    LatestPlotEnteringPhase2(),
    SetLatestPlotProgressForPhase2(),
    LatestPlotEnteringPhase3(),
    SetLatestPlotProgressForPhase3(),
    LatestPlotEnteringPhase4(),
    SetLatestPlotAsFinished(),
    StartCopyOfPlot(),
    # FinishedCopyOfPlot(),
)



import asyncio
import traceback
import uuid
from datetime import datetime
from typing import Dict, Tuple, Union

import grpc
from google.protobuf.json_format import MessageToDict

from ..chia_watchdog.ChiaWatchdog import ChiaWatchdog
from ..chia_watchdog.computer_info_comparison import compare_computer_info
from ..protobuf.data_collection.computer_info import collect_computer_info
from ..protobuf.generated.computer_info_pb2 import ComputerInfo, UpdateEvent
from ..protobuf.generated.config_pb2 import (
    _MONITORINGCONFIG_CLIENTCONFIG_SENDUPDATEEVERY,
    DevelopmentConfig,
    MonitoringConfig,
)
from ..protobuf.generated.monitoring_service_pb2 import DataUpdateRequest, GetStateRequest
from ..protobuf.generated.monitoring_service_pb2_grpc import MonitoringStub
from ..protobuf.to_sqlite.custom import ProtoType, get_update_even_data
from ..utils.logger import get_logger
from ..utils.settings import get_settings_value
from ..utils.timing import wait_at_least

ClientConfig = MonitoringConfig.ClientConfig


def load_machine_id() -> str:
    """Loads id of the machine

    Returns
    -------
    machine_id : str
        id of the machine

    Notes
    -----
        The machine id is stored in the home directory in
        `~/.chia_tea/settings.json`.
    """

    return get_settings_value("machineId", default=uuid.getnode())


def get_collection_frequencies(config: ClientConfig) -> Dict[str, float]:
    """Get the collection frequencies for updates to the server

    Parameters
    ----------
    config : ClientConfig
        config for the monitoring client

    Returns
    -------
    collection_frequencies : Dict[str, float]
        update event name and frequency pairs

    Raises
    ------
    ValueError
        In case an invalid name not matching the proto schema
        was given
    """

    collection_frequencies = config.send_update_every

    user_rate_limits = {}
    for field in _MONITORINGCONFIG_CLIENTCONFIG_SENDUPDATEEVERY.fields:
        field_value = getattr(collection_frequencies, field.name)
        user_rate_limits[field.name] = field_value

    update_event_var_names = tuple(
        field.name
        for field in UpdateEvent.DESCRIPTOR.fields
        if field.type == ProtoType.MESSAGE.value
    )

    for name, _ in user_rate_limits.items():
        if name not in update_event_var_names:
            err_msg = "Config entry '{0}' is not valid." + " Use one of {1}"
            raise ValueError(err_msg.format(name, ", ".join(update_event_var_names)))

    return user_rate_limits


class MonitoringClient:
    """Class for collecting and sending monitoring data to a server"""

    # pylint: disable=too-many-instance-attributes

    config: ClientConfig
    debug_config: DevelopmentConfig
    credentials_cert: str
    machine_id: str
    machine_name: str

    # throttling
    collection_frequencies: Dict[str, float]
    last_time_sent: Dict[Union[str, Tuple[str, str]], datetime]

    # watching stuff
    chia_dog: ChiaWatchdog

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        chia_dog: ChiaWatchdog,
        config=MonitoringConfig.ClientConfig,
        credentials_cert: str = "",
        machine_name: str = "",
    ):
        self.config = config
        self.credentials_cert = credentials_cert
        self.machine_id = load_machine_id()
        self.chia_dog = chia_dog
        self.collection_frequencies = get_collection_frequencies(config)
        self.last_time_sent = {}
        self.machine_name = machine_name

    def is_event_allowed_to_be_sent(self, pb_msg: UpdateEvent) -> bool:
        """Checks if a an update event is allowed to be sent

        Parameters
        ----------
        pb_msg : UpdateEvent
            event to be sent to the user

        Returns
        -------
        is_allowed : bool
            whether the event is allowed to be sent according to the
            data throttling limits
        """

        # get the name of the submessage to be updated
        field_name, sub_msg = get_update_even_data(pb_msg)
        field_id = getattr(sub_msg, "id") if hasattr(sub_msg, "id") else None

        field_key = field_name if field_id is None else (field_name, field_id)

        # no field name found, that is odd. Do a warning and
        # continue
        if not field_name:
            warn_msg = "Could not identify which field" + " was set in an update event: %s"
            get_logger(__file__).warning(warn_msg, MessageToDict(pb_msg))
            self.last_time_sent[field_key] = datetime.now()
            return True

        allowed_every = self.collection_frequencies.get(field_name)
        # no throttling active
        if allowed_every is None:
            self.last_time_sent[field_key] = datetime.now()
            return True

        last_time = self.last_time_sent.get(field_key)
        # not sent yet
        if last_time is None:
            self.last_time_sent[field_key] = datetime.now()
            return True

        # check if enough time passed
        is_allowed = (datetime.now() - last_time).total_seconds() > allowed_every

        # memorize if we are sending this one
        if is_allowed:
            self.last_time_sent[field_key] = datetime.now()

        return is_allowed

    async def __setup_channel(
        self,
        address: str,
    ) -> Tuple[grpc.aio.Channel, dict]:
        logger = get_logger(__name__)

        channel_args = {
            "target": address,
        }

        # add auth if specified
        if self.credentials_cert:
            logger.debug("Using authentication.")
            channel_args["credentials"] = grpc.ssl_channel_credentials(
                root_certificates=self.credentials_cert,
                # private_key=self.credentials_key,
            )
            channel_constructor = grpc.aio.secure_channel
        else:
            logger.debug("No authentication used.")
            channel_constructor = grpc.aio.insecure_channel

        return channel_constructor, channel_args

    async def send_infinite_update_requests(
        self,
        last_known_state: ComputerInfo,
        stub: MonitoringStub,
        address_for_logging: str,
    ):
        """Sends DataUpdateRequest in an infinite loop

        Parameters
        ----------
        last_known_state : ComputerInfo
            last known state to the server
        stub : MonitoringStub
            grpc stub connected to server
        address_for_logging : str
            ip address used for logging
        """
        logger = get_logger(__file__)

        stream = stub.SendMonitoringUpdate()

        previous_state = last_known_state
        while True:
            start_time = datetime.now()

            current_state = await collect_computer_info(
                self.machine_id,
                # we make a copy here, otherwise the object might get
                # mutated during data collection (takes a few ms).
                self.chia_dog.snapshot(),
            )

            event_list = [
                change_event
                async for change_event in compare_computer_info(
                    old_computer_info=previous_state, new_computer_info=current_state
                )
            ]

            filtered_event_list = [
                update_event
                for update_event in event_list
                if self.is_event_allowed_to_be_sent(update_event)
            ]
            if not filtered_event_list:
                await wait_at_least(
                    min_duration=self.config.collect_data_every, start_time=start_time
                )
                continue

            data_update_request = DataUpdateRequest(
                machine_id=self.machine_id,
                timestamp=datetime.now().timestamp(),
                events=filtered_event_list,
                machine_name=self.machine_name,
            )

            logger.info(
                "Sending message to %s: %s",
                address_for_logging,
                MessageToDict(data_update_request),
            )

            await stream.write(data_update_request)

            previous_state = current_state

            await wait_at_least(min_duration=self.config.collect_data_every, start_time=start_time)

    async def start_sending_updates(self):
        """Starts sending updates to the server"""

        logger = get_logger(__name__)
        logger.info("Starting to monitor system.")

        address = "{ip}:{port}".format(ip=self.config.address, port=self.config.port)

        channel_constructor, channel_args = await self.__setup_channel(
            address=address,
        )

        # don't do this at home kidz
        while True:
            try:
                # Open a connection to the server
                logger.debug("Connecting to %s", address)
                async with channel_constructor(**channel_args) as channel:

                    stub = MonitoringStub(channel)

                    # get last known state from the database
                    # we will compare to this and send the
                    # appropriate changes which happened
                    # in the meantime
                    logger.debug("Requesting last known state from server.")
                    last_known_state = await stub.GetMachineState(
                        GetStateRequest(machine_id=self.machine_id)
                    )
                    logger.debug(
                        "Received message %s",
                        MessageToDict(last_known_state),
                    )

                    # we can only send data once the watchdog
                    # is ready with it's init. Otherwise we will
                    # send updates for e.g. deleted plots which
                    # were just not found yet
                    await self.chia_dog.ready()

                    # send continuous updates
                    await self.send_infinite_update_requests(
                        last_known_state=last_known_state,
                        stub=stub,
                        address_for_logging=address,
                    )

            # in case of trouble reconnect
            except grpc.aio.AioRpcError as err:
                err_msg = "Connection error with %s (%s): %s"
                logger.error(
                    err_msg,
                    address,
                    str(err.code()),
                    err.details(),
                )
                await asyncio.sleep(5)

            except Exception:
                trace = traceback.format_exc()
                logger.error(trace)
                await asyncio.sleep(5)



from dataclasses import dataclass, field
from typing import List

from .FarmerHarvesterAPI import FarmerHarvesterAPI


@dataclass
class FarmerAPI:
    """This class holds chia information fetched through RPC
    from chia services on the same machine
    """

    # pylint: disable=too-few-public-methods

    is_ready: bool = False
    is_running: bool = False
    connections: List[FarmerHarvesterAPI] = field(default_factory=list)



import unittest

from google.protobuf.json_format import ParseDict

from ...protobuf.generated.computer_info_pb2 import ADD, DELETE, UPDATE, UpdateEvent
from .update_event_notifications import (
    notify_if_a_disk_is_lost,
    notify_if_plots_are_lost,
    notify_on_full_ram,
    notify_when_farmer_connects_or_disconnects_to_harvester,
)


class TestUpdateEventNotifications(unittest.TestCase):
    def test_msg_if_farmer_connects_to_harvesters(self):

        update_event = ParseDict(
            js_dict=dict(
                event_type=ADD,
                farmer_harvester=dict(
                    id="harvester id",
                    ip_address="some ip",
                ),
            ),
            message=UpdateEvent(),
        )

        messages = notify_when_farmer_connects_or_disconnects_to_harvester(
            machine_id="some machine id", update_events=[update_event]
        )

        self.assertEqual(len(messages), 1)

    def test_msg_if_farmer_disconnects_to_harvesters(self):

        update_event = ParseDict(
            js_dict=dict(
                event_type=DELETE,
                farmer_harvester=dict(
                    id="harvester id",
                    ip_address="some ip",
                ),
            ),
            message=UpdateEvent(),
        )

        messages = notify_when_farmer_connects_or_disconnects_to_harvester(
            machine_id="some machine id", update_events=[update_event]
        )

        self.assertEqual(len(messages), 1)

    def test_no_msg_if_farmer_to_harvesters_status_just_updates(self):

        update_event = ParseDict(
            js_dict=dict(
                event_type=UPDATE,
                farmer_harvester=dict(
                    id="harvester id",
                    ip_address="some ip",
                ),
            ),
            message=UpdateEvent(),
        )

        messages = notify_when_farmer_connects_or_disconnects_to_harvester(
            machine_id="some machine id", update_events=[update_event]
        )

        self.assertEqual(len(messages), 0)

    def test_msg_if_ram_full(self):

        update_event = ParseDict(
            js_dict=dict(
                ram=dict(
                    total_ram=100,
                    used_ram=96,
                )
            ),
            message=UpdateEvent(),
        )

        messages = notify_on_full_ram(
            machine_id="some machine id",
            update_events=[update_event],
        )

        self.assertEqual(len(messages), 1)

    def test_no_msg_if_ram_ok(self):

        update_event = ParseDict(
            js_dict=dict(
                ram=dict(
                    total_ram=100,
                    used_ram=80,
                )
            ),
            message=UpdateEvent(),
        )

        messages = notify_on_full_ram(
            machine_id="some machine id",
            update_events=[update_event],
        )

        self.assertEqual(len(messages), 0)

    def test_msg_if_disk_is_lost(self):

        update_event = ParseDict(
            js_dict=dict(
                event_type=DELETE,
                disk=dict(
                    name="/somedir",
                ),
            ),
            message=UpdateEvent(),
        )

        messages = notify_if_a_disk_is_lost("some machine id", update_events=[update_event])
        self.assertEqual(len(messages), 1)

    def test_no_msg_if_disk_is_added_or_updated(self):

        update_event = ParseDict(
            js_dict=dict(
                event_type=ADD,
                disk=dict(
                    name="/somedir",
                ),
            ),
            message=UpdateEvent(),
        )

        messages = notify_if_a_disk_is_lost("some machine id", update_events=[update_event])
        self.assertEqual(len(messages), 0)

        update_event.event_type = UPDATE
        messages = notify_if_a_disk_is_lost("some machine id", update_events=[update_event])
        self.assertEqual(len(messages), 0)

    def test_msg_if_plots_are_lost(self):

        update_event = ParseDict(
            js_dict=dict(
                event_type=DELETE,
                harvester_plot=dict(
                    id="some plot id 1",
                ),
            ),
            message=UpdateEvent(),
        )

        messages = notify_if_plots_are_lost("some machine id", update_events=[update_event])
        self.assertEqual(len(messages), 1)
        self.assertTrue("1" in messages[0])

    def test_no_msg_if_plot_is_added_or_updated(self):

        update_event = ParseDict(
            js_dict=dict(
                event_type=ADD,
                harvester_plot=dict(
                    id="some plot id 1",
                ),
            ),
            message=UpdateEvent(),
        )

        messages = notify_if_plots_are_lost("some machine id", update_events=[update_event])
        self.assertEqual(len(messages), 0)

        update_event.event_type = UPDATE
        messages = notify_if_plots_are_lost("some machine id", update_events=[update_event])
        self.assertEqual(len(messages), 0)



import unittest
from datetime import datetime

from ..ChiaWatchdog import ChiaWatchdog
from .FarmerHarvesterLogfile import FarmerHarvesterLogfile
from .line_checks import (
    ActionHarvesterConnected,
    ActionHarvesterDisconnected,
    ActionHarvesterFoundProof,
    ActionMessageFromHarvester,
    ActionFinishedSignagePoint,
    ActionMessageToHarvester,
)

# pylint: skip-file


class TestLineAction(unittest.TestCase):
    def test_harvester_disconnected(self):

        action = ActionHarvesterDisconnected()

        node_id = "d46fb9aaaa01f3aa3fc04f3e43231d35c3a1ddd4"
        ip_address = "57.22.39.97"
        timestamp_str = "2021-05-26T09:37:13.872"
        timestamp = datetime.fromisoformat(timestamp_str)

        line = (
            f"{timestamp_str} farmer farmer_server              : INFO"
            + f"     Connection closed: {ip_address},"
            + f" node id: {node_id}"
        )

        self.assertTrue(action.is_match(line))

        # check case that harvester does not exist yet
        chia_dog = ChiaWatchdog("", "")

        action.apply(line, chia_dog)
        self.assertTrue(len(chia_dog.harvester_infos) == 1)
        harvester_info = list(chia_dog.harvester_infos.values())[0]
        self.assertEqual(harvester_info.ip_address, ip_address)
        self.assertEqual(harvester_info.harvester_id, node_id)
        self.assertEqual(harvester_info.is_connected, False)
        self.assertEqual(harvester_info.last_update, timestamp)

        # check in case harvester exists
        chia_dog.harvester_infos = {
            node_id: FarmerHarvesterLogfile(
                node_id,
                ip_address,
                timestamp,
                True,
            )
        }
        action.apply(line, chia_dog)
        self.assertTrue(len(chia_dog.harvester_infos) == 1)
        harvester_info = list(chia_dog.harvester_infos.values())[0]
        self.assertEqual(harvester_info.ip_address, ip_address)
        self.assertEqual(harvester_info.harvester_id, node_id)
        self.assertEqual(harvester_info.is_connected, False)
        self.assertEqual(harvester_info.last_update, timestamp)

    def test_harvester_connected(self):

        action = ActionHarvesterConnected()

        node_id = "d46fb9aaaa01f3aa3fc04f3e43231d35c3a1ddd4"
        ip_address = "57.22.39.97"
        timestamp_str = "2021-05-26T09:37:13.872"
        timestamp = datetime.fromisoformat(timestamp_str)

        line = (
            f"{timestamp_str} farmer farmer_server           "
            + f"   : DEBUG    -> harvester_handshake to peer {ip_address} "
            + node_id
        )

        self.assertTrue(action.is_match(line))

        # check case that harvester does not exist yet
        chia_dog = ChiaWatchdog("", "")

        action.apply(line, chia_dog)
        self.assertTrue(len(chia_dog.harvester_infos) == 1)
        harvester_info = list(chia_dog.harvester_infos.values())[0]
        self.assertEqual(harvester_info.ip_address, ip_address)
        self.assertEqual(harvester_info.harvester_id, node_id)
        self.assertEqual(harvester_info.is_connected, True)
        self.assertEqual(harvester_info.last_update, timestamp)

        # check in case harvester exists
        chia_dog.harvester_infos = {
            node_id: FarmerHarvesterLogfile(
                node_id,
                ip_address,
                timestamp,
                True,
            )
        }
        action.apply(line, chia_dog)
        self.assertTrue(len(chia_dog.harvester_infos) == 1)
        harvester_info = list(chia_dog.harvester_infos.values())[0]
        self.assertEqual(harvester_info.ip_address, ip_address)
        self.assertEqual(harvester_info.harvester_id, node_id)
        self.assertEqual(harvester_info.is_connected, True)
        self.assertEqual(harvester_info.last_update, timestamp)

    def test_farmer_message_to_harvester(self):

        action = ActionMessageToHarvester()

        node_id = "d46fb9aaaa01f3aa3fc04f3e43231d35c3a1ddd4"
        ip_address = "57.22.39.97"
        timestamp_str = "2021-05-26T09:37:13.872"

        line = msg_to_harvester(timestamp_str, ip_address, node_id)
        # check line detection
        self.assertTrue(action.is_match(line))

        # check correct modification
        chia_dog = ChiaWatchdog("", "")
        action.apply(line, chia_dog)

        self.assertEqual(len(chia_dog.harvester_infos), 1)
        harvester = chia_dog.harvester_infos[node_id]
        self.assertEqual(harvester.ip_address, ip_address)
        self.assertEqual(harvester.harvester_id, node_id)
        self.assertEqual(harvester.is_connected, False)

    def test_farmer_message_from_harvester(self):

        action = ActionMessageFromHarvester()

        node_id = "d46fb9aaaa01f3aa3fc04f3e43231d35c3a1ddd4"
        ip_address = "57.22.39.97"
        timestamp_str = "2021-05-26T09:37:13.872"

        line = msg_from_harvester(timestamp_str, ip_address, node_id)

        # check line detection
        self.assertTrue(action.is_match(line))

        # check correct modification
        chia_dog = ChiaWatchdog("", "")
        action.apply(line, chia_dog)

        self.assertEqual(len(chia_dog.harvester_infos), 1)
        harvester = chia_dog.harvester_infos[node_id]
        self.assertEqual(harvester.ip_address, ip_address)
        self.assertEqual(harvester.harvester_id, node_id)
        self.assertEqual(harvester.is_connected, True)

    def test_harvester_found_proof(self):
        action1 = ActionHarvesterFoundProof()

        line = (
            "2021-06-16T22:44:56.890 "
            + "harvester chia.harvester.harvester: "
            + "INFO 0 plots were eligible for farming 65322a31ad... "
            + "Found 0 proofs. Time: 0.00015 s. Total 0 plots"
        )
        line1Found = (
            "2021-06-16T22:44:56.890 "
            + "harvester chia.harvester.harvester: "
            + "INFO 0 plots were eligible for farming 65322a31ad... "
            + "Found 1 proofs. Time: 0.00015 s. Total 0 plots"
        )
        line10Found = (
            "2021-06-16T22:44:56.890 "
            + "harvester chia.harvester.harvester: "
            + "INFO 0 plots were eligible for farming 65322a31ad... "
            + "Found 10 proofs. Time: 0.00015 s. Total 0 plots"
        )

        self.assertTrue(action1.is_match(line))
        self.assertTrue(action1.is_match(line1Found))
        self.assertTrue(action1.is_match(line10Found))

        chia_dog = ChiaWatchdog("", "")

        action1.apply(line, chia_dog)
        harvester_service = chia_dog.harvester_service
        self.assertEqual(harvester_service.n_proofs, 0)
        action1.apply(line1Found, chia_dog)
        self.assertEqual(harvester_service.n_proofs, 1)
        action1.apply(line10Found, chia_dog)
        self.assertEqual(harvester_service.n_proofs, 11)

    def test_harvester_msgs_timouts(self):
        """
        This test is done for the new logic with sgn points
        """
        # pylint: disable=too-many-locals

        chia_dog = ChiaWatchdog("", "")

        node_id = "d46fb9aaaa01f3aa3fc04f3e43231d35c3a1ddd4"
        ip_address = "57.22.39.97"

        timestamp0_str = "2021-05-26T09:30:00.872"  # signage - should not fail
        timestamp1_str = "2021-05-26T09:30:01.872"  # send
        timestamp2_str = "2021-05-26T09:30:02.872"  # recieve
        timestamp3_str = "2021-05-26T09:30:20.872"  # signage - not timed out

        timestamp4_str = "2021-05-26T09:30:21.872"  # send
        timestamp5_str = "2021-05-26T09:31:51.872"  # signage - timeout

        timestamp6_str = "2021-05-26T09:44:52.872"  # recieve - (connect)

        line0 = msg_signage_point(timestamp0_str)
        line1 = msg_to_harvester(timestamp1_str, ip_address, node_id)
        line2 = msg_from_harvester(timestamp2_str, ip_address, node_id)
        line3 = msg_signage_point(timestamp3_str)

        ActionFinishedSignagePoint().apply(line0, chia_dog)
        ActionMessageToHarvester().apply(line1, chia_dog)
        ActionMessageFromHarvester().apply(line2, chia_dog)
        ActionFinishedSignagePoint().apply(line3, chia_dog)
        harvester_info = chia_dog.harvester_infos[node_id]
        self.assertFalse(harvester_info.timed_out)

        line4 = msg_to_harvester(timestamp4_str, ip_address, node_id)
        line5 = msg_signage_point(timestamp5_str)
        ActionMessageToHarvester().apply(line4, chia_dog)
        ActionFinishedSignagePoint().apply(line5, chia_dog)
        self.assertTrue(harvester_info.timed_out)
        self.assertEqual(harvester_info.n_overdue_responses, 1)

        line6 = msg_from_harvester(timestamp6_str, ip_address, node_id)
        ActionMessageFromHarvester().apply(line6, chia_dog)
        self.assertFalse(harvester_info.timed_out)

    def test_harvester_disconnect_times_reset_and_timeout(self):
        """
        This test is for the negative times which occured.
        Events:
        - send
        - disconnect XXX
        - send
        - send
        - connect XXX
        - recieve
        - recieve
        - send
        - recieve
        """
        # pylint: disable=too-many-locals

        actionOut = ActionMessageToHarvester()
        actionIn = ActionMessageFromHarvester()

        actionConnect = ActionHarvesterConnected()
        actionDisconnect = ActionHarvesterDisconnected()

        node_id = "d46fb9aaaa01f3aa3fc04f3e43231d35c3a1ddd4"
        ip_address = "57.22.39.97"
        timestamp1_str = "2021-05-26T09:38:01.872"  # send
        timestamp2_str = "2021-05-26T09:38:02.872"  # disconnect

        timestamp3_str = "2021-05-26T09:38:41.872"  # send

        timestamp4_str = "2021-05-26T09:38:56.872"  # connect

        timestamp5_str = "2021-05-26T09:43:39.872"  # recieve
        timestamp6_str = "2021-05-26T09:44:39.872"  # send
        timestamp7_str = "2021-05-26T09:45:39.872"  # recieve

        timestamp = datetime.fromisoformat(timestamp1_str)

        # pylint: disable=duplicate-code
        chia_dog = ChiaWatchdog("", "")
        chia_dog.harvester_infos = {
            node_id: FarmerHarvesterLogfile(
                node_id,
                ip_address,
                True,
                timestamp,
            )
        }

        harvester_info = list(chia_dog.harvester_infos.values())[0]

        # Here the testingbegins
        # send a challenge
        lineOut = msg_to_harvester(timestamp1_str, ip_address, node_id)
        actionOut.apply(lineOut, chia_dog)

        # disconnect
        lineDisconnect = msg_disconnect(timestamp2_str, ip_address, node_id)
        actionDisconnect.apply(lineDisconnect, chia_dog)

        self.assertFalse(harvester_info.is_connected)

        # send - late
        line = msg_signage_point(timestamp3_str)
        ActionFinishedSignagePoint().apply(line, chia_dog)
        lineOut = msg_to_harvester(timestamp3_str, ip_address, node_id)
        actionOut.apply(lineOut, chia_dog)
        self.assertFalse(harvester_info.is_connected)

        # connect
        lineConnect = msg_connect(timestamp4_str, ip_address, node_id)
        actionConnect.apply(lineConnect, chia_dog)
        self.assertTrue(harvester_info.is_connected)
        self.assertFalse(harvester_info.timed_out)

        # recieve  - late
        lineIn = msg_from_harvester(timestamp5_str, ip_address, node_id)
        actionIn.apply(lineIn, chia_dog)

        # send recieve
        lineOut = msg_to_harvester(timestamp6_str, ip_address, node_id)
        lineIn = msg_from_harvester(timestamp7_str, ip_address, node_id)

        actionOut.apply(lineOut, chia_dog)
        actionIn.apply(lineIn, chia_dog)


def msg_to_harvester(timestamp_str: str, ip_address: str, node_id: str) -> str:
    """Get a fake log msg indicating a send msg to harvester"""
    line = (
        f"{timestamp_str} farmer farmer_server              : DEBUG"
        + f"    -> new_signage_point_harvester to peer {ip_address} {node_id}"
    )
    return line


def msg_from_harvester(timestamp_str: str, ip_address: str, node_id: str) -> str:
    """Get a fake log msg when a farmer receives data from a harvester"""
    line = (
        f"{timestamp_str} farmer farmer_server              : DEBUG"
        + f"    <- farming_info from peer {node_id} {ip_address}"
    )
    return line


def msg_signage_point(timestamp_str: str) -> str:
    """Get a fake log msg in case a new signage point has started"""
    line = (
        f"{timestamp_str} full_node chia.full_node.full_node: INFO"
        + ":timer:  Finished signage point 19/64: CC: RC:"
    )
    return line


def msg_disconnect(timestamp_str: str, ip_address: str, node_id: str) -> str:
    """Get a fake log msg for a disconnecting harvester to a farmer"""
    line = (
        f"{timestamp_str} farmer farmer_server              : INFO"
        + f"     Connection closed: {ip_address},"
        + f" node id: {node_id}"
    )
    return line


def msg_connect(timestamp_str: str, ip_address: str, node_id: str) -> str:
    """Get a fake log msg for a harvester connecting to a farmer"""
    line = (
        f"{timestamp_str} farmer farmer_server           "
        + f"   : DEBUG    -> harvester_handshake to peer {ip_address} "
        + node_id
    )
    return line



import traceback
from abc import ABC, abstractmethod
from datetime import datetime

from ...utils.logger import get_logger
from ..ChiaWatchdog import ChiaWatchdog


class AbstractLineAction(ABC):
    """Action to be applied on log lines"""

    @abstractmethod
    def is_match(self, line: str) -> bool:
        """Checks if a line is matching the Action

        Parameter
        ---------
        line : str
            line to be checked for a match

        Returns
        -------
        is_match : bool
            if the line is a match
        """
        raise NotImplementedError()

    @abstractmethod
    def apply(
        self: "AbstractLineAction",
        line: str,
        chia_dog: ChiaWatchdog,
    ):
        """Apply the line action

        Parameter
        ---------
        line : str
            line from which information will be extracted
        chia_dog : ChiaWatchdog
            chia watchdog to be modified
        """
        raise NotImplementedError()


class ActionMessageFromHarvester(AbstractLineAction):
    """This action is triggered if a farmer sends a msg to a harvester"""

    def is_match(self, line: str) -> bool:
        codewords = ("farmer farmer_server", "<-", "farming_info from peer")
        if all(word in line for word in codewords):
            return True
        return False

    def apply(
        self,
        line: str,
        chia_dog: ChiaWatchdog,
    ):

        fragments = line.split()

        # extract data from line
        timestamp = fragments[0]
        timestamp_dt = datetime.fromisoformat(timestamp)
        harvester_id = fragments[-2]
        ip_address = fragments[-1]

        # update info of harvester
        harvester_info = chia_dog.get_or_create_harvester_info(harvester_id, ip_address)

        # First out then in
        harvester_info.time_last_incoming_msg = timestamp_dt
        harvester_info.last_update = timestamp_dt
        harvester_info.is_connected = True
        harvester_info.timed_out = False
        chia_dog.harvester_infos[harvester_id] = harvester_info


class ActionMessageToHarvester(AbstractLineAction):
    """This action is triggered if a farmer sends a msg to a harvester"""

    def is_match(self, line: str) -> bool:
        codewords = ("farmer farmer_server", "->", "new_signage_point_harvester")
        if all(word in line for word in codewords):
            return True
        return False

    def apply(
        self,
        line: str,
        chia_dog: ChiaWatchdog,
    ):
        fragments = line.split()

        # extract data from line
        timestamp = fragments[0]
        timestamp_dt = datetime.fromisoformat(timestamp)
        harvester_id = fragments[-1]
        ip_address = fragments[-2]

        # update info of harvester
        harvester_info = chia_dog.get_or_create_harvester_info(harvester_id, ip_address)

        harvester_info.time_last_outgoing_msg = timestamp_dt
        harvester_info.last_update = timestamp_dt
        chia_dog.harvester_infos[harvester_id] = harvester_info


class ActionFinishedSignagePoint(AbstractLineAction):
    """
    Action is currently used as check if a Harvester is timed out.
    Might be used for SignPoint Metrics at a later stage
    """

    def is_match(self, line: str) -> bool:
        codewords = ("full_node", "Finished signage point")
        if all(word in line for word in codewords):
            return True
        return False

    def apply(
        self,
        line: str,
        chia_dog: ChiaWatchdog,
    ):
        fragments = line.split()

        # extract data from line
        timestamp = fragments[0]
        timestamp_dt = datetime.fromisoformat(timestamp)

        # Harvester Time out Check
        for farmer_harvester_logfile in chia_dog.harvester_infos.values():
            if farmer_harvester_logfile.is_connected:
                farmer_harvester_logfile.check_for_timeout(timestamp_dt)


class ActionHarvesterConnected(AbstractLineAction):
    """This action is triggered if a farmer connects to a harvester"""

    def is_match(self, line: str) -> bool:
        codewords = ("farmer farmer_server", "harvester_handshake to peer")
        if all(word in line for word in codewords):
            return True
        return False

    def apply(
        self,
        line: str,
        chia_dog: ChiaWatchdog,
    ):
        fragments = line.split()

        # extract data from line
        timestamp_dt = datetime.fromisoformat(fragments[0])
        harvester_id = fragments[-1]
        ip_address = fragments[-2]

        # update info of harvester
        harvester_info = chia_dog.get_or_create_harvester_info(harvester_id, ip_address)
        harvester_info.is_connected = True
        harvester_info.timed_out = False
        harvester_info.last_update = timestamp_dt
        chia_dog.harvester_infos[harvester_id] = harvester_info


class ActionHarvesterDisconnected(AbstractLineAction):
    """This action is triggered if a farmer disconnects to a harvester"""

    def is_match(self, line: str) -> bool:
        codewords = ("farmer farmer_server", "Connection closed")
        if all(word in line for word in codewords):
            return True
        return False

    def apply(
        self,
        line: str,
        chia_dog: ChiaWatchdog,
    ):
        fragments = line.split()

        # extract data from line
        timestamp_dt = datetime.fromisoformat(fragments[0])
        harvester_id = fragments[-1]
        ip_address = fragments[-4]

        # remove suffix
        if ip_address.endswith(","):
            ip_address = ip_address[:-1]

        # update info of harvester
        harvester_info = chia_dog.get_or_create_harvester_info(harvester_id, ip_address)
        harvester_info.is_connected = False
        harvester_info.last_update = timestamp_dt
        chia_dog.harvester_infos[harvester_id] = harvester_info


class ActionHarvesterFoundProof(AbstractLineAction):
    """This action is triggered if a harvester found a proof"""

    # Chia Version: 1.2.0
    # Example:
    #
    # 2021-07-17T13:34:34.513 harvester chia.harvester.harvester: INFO
    # 0 plots were eligible for farming 142fd5714f...
    # Found 0 proofs. Time: 0.00017 s. Total 0 plots

    def is_match(self, line: str) -> bool:
        codewords = (
            "harvester chia.harvester.harvester",
            "eligible",
            "Found",
            "proofs",
        )
        if all(word in line for word in codewords):
            return True
        return False

    def apply(
        self,
        line: str,
        chia_dog: ChiaWatchdog,
    ):
        fragments = line.split()

        # extract data from line
        proofs = int(fragments[-8])  # interger number

        if proofs > 0:
            harvester_service = chia_dog.harvester_service
            harvester_service.n_proofs += proofs


class ActionFarmedUnfinishedBlock(AbstractLineAction):
    """This action is triggered if a harvester found a block"""

    def is_match(self, line: str) -> bool:
        codewords = ("full_node chia.full_node.full_node", "Farmed unfinished_block")
        if all(word in line for word in codewords):
            return True
        return False

    def apply(
        self,
        line: str,
        chia_dog: ChiaWatchdog,
    ):
        fragments = line.split()

        # extract block id from fragements
        block_id = fragments[-1]
        if block_id not in chia_dog.farmed_blocks:
            chia_dog.farmed_blocks.append(block_id)


async def run_line_checks(chia_dog: ChiaWatchdog, line: str):
    """Processes a line from the logfile

    Parameters
    ----------
    line : str
        logfile line
    """
    try:
        if line:
            for action in ALL_LINE_ACTIONS:
                if action.is_match(line):
                    action.apply(line, chia_dog)

    except Exception:
        trace = traceback.format_exc()
        err_msg = "Error in line: %s\n%s"
        get_logger(__name__).error(err_msg, line, trace)


ALL_LINE_ACTIONS = (
    ActionMessageFromHarvester(),
    ActionMessageToHarvester(),
    ActionFinishedSignagePoint(),
    ActionHarvesterConnected(),
    ActionHarvesterDisconnected(),
    ActionHarvesterFoundProof(),
)



import asyncio
import logging
import sqlite3
import traceback
from datetime import datetime
from typing import Dict, List, Tuple

import discord

from ...protobuf.generated.computer_info_pb2 import ComputerInfo
from ...protobuf.generated.machine_info_pb2 import MachineInfo
from ...protobuf.to_sqlite.sql_cmds import (
    get_computer_info_from_db,
    get_machine_infos_from_db,
    get_update_events_from_db,
)
from ...utils.logger import get_logger
from ..common import open_database_read_only
from .computer_info_notifications import get_computer_info_messages_if_any
from .update_event_notifications import get_update_event_messages_if_any

DISCORD_MSG_LIMIT = 2000
MSG_TOO_LONG = "⚠️  Message too long for discord ({n_chars_too_long} > {discord_msg_limit} chars)"


async def log_and_send_msg_if_any(
    messages: List[str],
    logger: logging.Logger,
    channel: discord.ChannelType,
    is_testing: bool,
):
    """Logs and sends the messages if not empty

    Parameters
    ----------
    messages : List[str]
        messages to print
    logger : logging.Logger
        logger for printing
    channel : discord.ChannelType
        discord channel to send messages to
    is_testing : bool
        whether to run in testing mode and not
        send but log messages
    """

    if messages:

        # Send all messages as once piece if we don't hit
        # discords API limit.
        total_msg = "\n".join(messages)
        if len(total_msg) < DISCORD_MSG_LIMIT:
            messages = [total_msg]

        for msg in messages:
            n_chars_too_long = len(msg) - DISCORD_MSG_LIMIT
            if n_chars_too_long > 0:
                msg = MSG_TOO_LONG.format(
                    n_chars_too_long=n_chars_too_long, discord_msg_limit=DISCORD_MSG_LIMIT
                )

            logger.info(msg)
            if not is_testing:
                await channel.send(msg)


def get_current_computer_and_machine_infos_from_db(
    cursor: sqlite3.Cursor,
) -> Dict[str, Tuple[MachineInfo, ComputerInfo]]:
    """Get all current computer infos from the database

    Parameters
    ----------
    cursor: sqlite3.Cursor
        cursor to the sqlite3 db

    Returns
    -------
    computer_and_machine_infos : Dict[str, Tuple[MachineInfo, ComputerInfo]]
        dict with machine and computer info as value and machine
        id as key
    """
    machine_info_list, _ = get_machine_infos_from_db(cursor)

    computer_and_machine_infos = {}
    for machine in machine_info_list:
        computer_and_machine_infos[machine.machine_id] = (
            machine,
            get_computer_info_from_db(
                cursor,
                machine.machine_id,
            ),
        )

    return computer_and_machine_infos


async def run_notifiers(
    db_filepath: str,
    channel: discord.ChannelType,
    is_testing: bool,
):
    """Notify channel in case a harvester got lost

    Parameters
    ----------
    db_filepath : str
        path to the monitoring db
    channel : discord.ChannelType
        discord channel for notifications
    is_testing : bool
        whether to run in testing mode and not
        send but log messages
    """
    # pylint: disable=too-many-locals

    logger = get_logger(__name__)

    # time for a looping
    while True:
        try:

            with open_database_read_only(db_filepath) as cursor:

                # fetch the state of all machines last known
                old_machine_computer_info_dict = get_current_computer_and_machine_infos_from_db(
                    cursor
                )

                last_timestamp = int(datetime.now().timestamp())
                while True:

                    # wait inbetween status checks, no need to go crazy
                    # also kind of nice since the bot does not enrage
                    wait_time_between_checks = 2
                    await asyncio.sleep(wait_time_between_checks)

                    new_timestamp = int(datetime.now().timestamp())
                    messages = []

                    # messages related to update events
                    machine_events = get_update_events_from_db(
                        cursor,
                        last_timestamp,
                        new_timestamp,
                    )

                    for machine_id, event_list in machine_events.items():
                        messages += get_update_event_messages_if_any(
                            machine_id,
                            event_list,
                        )

                    # messages by computer info comparison
                    new_machine_computer_info_dict = get_current_computer_and_machine_infos_from_db(
                        cursor,
                    )

                    for machine_id, (
                        machine_info,
                        new_computer_info,
                    ) in new_machine_computer_info_dict.items():
                        _, old_computer_info = old_machine_computer_info_dict.get(machine_id)
                        new_computer_info = get_computer_info_from_db(
                            cursor,
                            machine_id,
                        )
                        messages += get_computer_info_messages_if_any(
                            machine=machine_info,
                            old_computer_info=old_computer_info,
                            new_computer_info=new_computer_info,
                        )

                    # print message if any
                    await log_and_send_msg_if_any(
                        messages=messages,
                        logger=logger,
                        channel=channel,
                        is_testing=is_testing,
                    )

                    # if all went fine we can update the last timestamp
                    last_timestamp = new_timestamp
                    # and the last known state
                    old_machine_computer_info_dict = new_machine_computer_info_dict

        # something on db level failed
        except Exception:
            tb = traceback.format_exc()
            get_logger(__file__).error(tb)

            await asyncio.sleep(2)



import os
import sys

from discord.ext import commands

from chia_tea.discord.commands.plotters import plotters_cmd

from ..utils.cli import parse_args
from ..utils.config import get_config, read_config
from ..utils.logger import get_logger
from .commands.farmers import farmers_cmd
from .commands.harvesters import harvesters_cmd
from .commands.machines import machines_cmd
from .commands.sql import sql_cmd
from .commands.wallets import wallets_cmd
from .notifications.run_notifiers import log_and_send_msg_if_any, run_notifiers

# setup log handler
module_name = os.path.splitext(os.path.basename(__file__))[0]

# =========================
#         DISCORD
# =========================

bot = commands.Bot(command_prefix="$")
channel_id = ""


@bot.command(name="hi")
async def bot_hi(ctx):
    """Says hi to the user"""
    await ctx.send("Hi!")


@bot.command(name="wallets")
async def bot_wallets(ctx):
    """Prints all wallets"""
    db_filepath = get_config().monitoring.server.db_filepath

    messages = await wallets_cmd(db_filepath)

    await log_and_send_msg_if_any(
        messages=messages,
        logger=get_logger(__file__),
        channel=ctx.channel,
        is_testing=get_config().development.testing,
    )


@bot.command(name="machines")
async def bot_machines(ctx):
    """Prints all machines"""
    db_filepath = get_config().monitoring.server.db_filepath

    messages = await machines_cmd(db_filepath)

    await log_and_send_msg_if_any(
        messages=messages,
        logger=get_logger(__file__),
        channel=ctx.channel,
        is_testing=get_config().development.testing,
    )


@bot.command(name="farmers")
async def bot_farmers(ctx):
    """Prints all farmers"""
    db_filepath = get_config().monitoring.server.db_filepath

    messages = await farmers_cmd(db_filepath)

    await log_and_send_msg_if_any(
        messages=messages,
        logger=get_logger(__file__),
        channel=ctx.channel,
        is_testing=get_config().development.testing,
    )


@bot.command(name="harvesters")
async def bot_harvester(ctx):
    """Prints all harvesters"""
    db_filepath = get_config().monitoring.server.db_filepath

    messages = await harvesters_cmd(db_filepath)

    await log_and_send_msg_if_any(
        messages=messages,
        logger=get_logger(__file__),
        channel=ctx.channel,
        is_testing=get_config().development.testing,
    )


@bot.command(name="plotters")
async def bot_plotters(ctx):
    """Prints all plotters"""
    db_filepath = get_config().monitoring.server.db_filepath

    messages = await plotters_cmd(db_filepath)

    await log_and_send_msg_if_any(
        messages=messages,
        logger=get_logger(__file__),
        channel=ctx.channel,
        is_testing=get_config().development.testing,
    )


@bot.command(name="sql")
async def bot_sql(ctx, *cmds):
    """Let's the user execute arbitrary sql statements"""
    db_filepath = get_config().monitoring.server.db_filepath

    cmd = " ".join(cmds)
    messages = await sql_cmd(db_filepath, cmd)

    await log_and_send_msg_if_any(
        messages=messages,
        logger=get_logger(__file__),
        channel=ctx.channel,
        is_testing=get_config().development.testing,
    )


def get_discord_channel_id() -> int:
    """Get the discord channel id from the config

    Returns
    -------
    channel_id_cfg : int
        discord channel id

    Raises
    ------
    ValueError
        In case that the channel id cannot be converte to a string
    """
    channel_id_cfg = get_config().discord.channel_id

    # yaml sometimes returns a str depending on how
    # the user put the value in, but we need an int
    if not isinstance(channel_id_cfg, int):
        try:
            channel_id_cfg = int(channel_id_cfg)
        except ValueError:
            err_msg = "Cannot convert discord channel id '{0}' to number."
            get_logger(module_name).error(err_msg, channel_id_cfg)
            sys.exit(1)

    return channel_id_cfg


@bot.event
async def on_ready():
    """Function run when the bot is ready"""
    db_filepath = get_config().monitoring.server.db_filepath

    get_logger(module_name).info("Bot started.")

    # get discord channel
    channel = bot.get_channel(channel_id)

    config = get_config()
    is_testing = config.development.testing

    await run_notifiers(db_filepath, channel, is_testing)


def main():
    """Main function for the discord bot"""
    args = parse_args(
        name="Chia Tea Discord Bot",
        description="Start a discord bot to keep an eye on your Chia farm.",
    )

    # load config
    config = read_config(args.config)

    discord_token = config.discord.token

    # pylint: disable=global-statement
    global channel_id
    channel_id = get_discord_channel_id()

    # initiate bot and event loop
    # create_task must come first!!!!
    bot.run(discord_token)


if __name__ == "__main__":
    main()



import inspect
import logging
import os
import sys
import traceback
from functools import wraps
from logging import StreamHandler
from time import time
from typing import Callable

from concurrent_log_handler import ConcurrentRotatingFileHandler

from ..protobuf.generated.config_pb2 import DEBUG, ERROR, INFO, TRACE, WARNING
from .config import get_config

TRACE_LEVEL = 5
TRACE_NAME = "TRACE"
logging.addLevelName(TRACE_LEVEL, TRACE_NAME)


def __parse_loglevel(level: str) -> int:
    """Parse the log leve froms string into integer

    Parameters
    ----------
    level : str
        one of: DEBUG, INFO, WARNING, ERROR

    Returns
    -------
    loglevel : int
        log leve as int for the logging module
    """
    if level == INFO:
        return logging.INFO
    if level == DEBUG:
        return logging.DEBUG
    if level == WARNING:
        return logging.WARNING
    if level == ERROR:
        return logging.ERROR
    if level == TRACE:
        return TRACE_LEVEL

    msg = (
        f"Name '{0}' is not an a correct log level."
        "Try one of TRACE, DEBUG, INFO, WARNING, ERROR."
    )
    raise ValueError(msg.format(level))


def __add_handler_only_once(logger: logging.Logger, handler: logging.Handler, handler_type: type):
    """Adds a handler only if there is none of the same class yet

    Parameters
    ----------
    logger : logging.Logger
        logger to add handler to
    handler : logging.Handler
        handler to add
    """
    if not any(isinstance(handler, handler_type) for handler in logger.handlers):
        logger.addHandler(handler)


def get_logger(
    name: str,
) -> logging.Logger:
    """Get a logger with a (module) name

    Parameters
    ----------
    name : str
        name of the logger, preferably '__name__' for modules

    Returns
    -------
    logger : logging.Logger
    """
    new_logger = logging.getLogger(name)
    config = get_config().logging

    # print to stdout
    #
    # must be BEFORE file handler otherwise it will
    # not be added due to the logger type check
    if config.log_to_console:
        __add_handler_only_once(
            new_logger,
            StreamHandler(stream=sys.stdout),
            StreamHandler,
        )

    # log to file if possible
    try:
        # setup log handler (file)
        directory: str = os.path.join(
            os.path.expanduser("~/.chia_tea/log"),
        )
        os.makedirs(directory, exist_ok=True)

        if config.log_to_file:
            handler = ConcurrentRotatingFileHandler(
                os.path.join(directory, "chia_tea.log"),
                maxBytes=config.max_logfile_size_mb * 1024 * 1024,
                backupCount=config.max_logfiles,
            )
            formatter = logging.Formatter(
                fmt="%(asctime)s %(levelname)-8s %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
            handler.setFormatter(formatter)
            __add_handler_only_once(new_logger, handler, ConcurrentRotatingFileHandler)

    # pylint: disable=broad-except
    except Exception:
        err_msg = traceback.format_exc()
        logging.getLogger(__name__).error(err_msg)

    # loglevel
    loglevel_str = config.loglevel
    loglevel = __parse_loglevel(loglevel_str)
    new_logger.setLevel(loglevel)

    return new_logger


def get_function_name(function: Callable) -> str:
    """Get the proper name of a function as string

    Parameters
    ----------
    function : Callable
        function to get the name from

    Returns
    -------
    name : str
    """
    name = function.__module__ + "." + function.__name__

    for cls in inspect.getmro(function.__class__):
        if function.__name__ in cls.__dict__:
            name += cls.__name__ + name

    return name


def log_runtime_async(module_name: str):
    """Logs the time of the function call

    Parameters
    ----------
    func : Callable
    """

    def decorator(function):
        @wraps(function)
        async def _time_it(*args, **kwargs):
            start = time()
            try:
                return await function(*args, **kwargs)
            finally:
                get_logger(module_name).log(
                    TRACE_LEVEL,
                    "{0} took {1:.2f}s".format(
                        get_function_name(function),
                        time() - start,
                    ),
                )

        return _time_it

    return decorator



import copy
import asyncio
from typing import Dict, List

from .api.FarmerAPI import FarmerAPI
from .api.HarvesterAPI import HarvesterAPI
from .api.WalletAPI import WalletAPI
from .logfile.FarmerHarvesterLogfile import FarmerHarvesterLogfile
from .madmax.MadMaxPlotInProgress import MadMaxPlotInProgress


class ChiaWatchdog:
    """Class for watching chia"""

    # pylint: disable=too-many-instance-attributes

    __logfile_chia_ready: bool = False
    __logfile_madmax_ready: bool = False

    # members related to logfile checking
    harvester_infos: Dict[str, FarmerHarvesterLogfile]
    farmed_blocks: List[str]

    # members related to madmax plotter
    plots_in_progress: List[MadMaxPlotInProgress]

    # members for contacting chia directly
    farmer_service: FarmerAPI
    wallet_service: WalletAPI
    harvester_service: HarvesterAPI

    def __init__(self, logfile_filepath: str, madmax_logfile: str):
        """initialize a chia watchdog

        Parameters
        ----------
        logfile_filepath : str
            path to the logfile to watch
        madmax_logfile : str
            path to the madmax logfile to watch
        """
        self.logfile_filepath = logfile_filepath
        self.madmax_logfile = madmax_logfile
        self.harvester_infos = {}
        self.farmer_service = FarmerAPI()
        self.wallet_service = WalletAPI()
        self.harvester_service = HarvesterAPI()
        self.farmed_blocks = []
        self.plots_in_progress = []

    async def ready(self):
        """Wait for the readiness of the watchdog"""
        while not (
            self.__logfile_chia_ready
            and self.__logfile_madmax_ready
            and self.harvester_service.is_ready
            and self.farmer_service.is_ready
            and self.wallet_service.is_ready
        ):
            await asyncio.sleep(0.25)

    def set_chia_logfile_is_ready(self):
        """When the chia logfile scanner has done its init, this gets called"""
        self.__logfile_chia_ready = True

    def set_madmax_logfile_is_ready(self):
        """When the madmax logfile scanner has done its init, this gets called"""
        self.__logfile_madmax_ready = True

    def get_or_create_harvester_info(
        self,
        harvester_id: str,
        ip_address: str,
    ) -> FarmerHarvesterLogfile:
        """Get or create a harvester info

        Parameters
        ----------
        harvester_id : str
            id of the harvester
        ip_address : str
            ip address

        Returns
        -------
        harvester_info : FarmerHarvesterLogfile
            existing or newly created harvester info
        """

        harvester_info = self.harvester_infos.get(harvester_id)

        #  does not exist yet
        if harvester_info is None:
            harvester_info = FarmerHarvesterLogfile(
                harvester_id=harvester_id,
                ip_address=ip_address,
            )
            self.harvester_infos[harvester_id] = harvester_info
        else:
            # make sure ip is updated
            harvester_info.ip_address = ip_address

        return harvester_info

    def snapshot(self) -> "ChiaWatchdog":
        """Takes a snapshot of the current watchdog

        Returns
        -------
        snapshot : ChiaWatchdog
            copy snapshot
        """
        return copy.deepcopy(self)



import asyncio
from typing import Callable


def async_test(fun: Callable):
    """Decorator for running async tests

    Parameters
    ----------
    f : function
        async test function to wrap
    """

    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(fun(*args, **kwargs))

    return wrapper



import asyncio

from ..chia_watchdog.ChiaWatchdog import ChiaWatchdog
from ..chia_watchdog.run_watchdog import run_watchdog
from ..utils.cli import parse_args
from ..utils.config import read_config
from .common import get_credentials_cert
from .MonitoringClient import MonitoringClient


def main():
    """This function starts the chia tea client"""
    args = parse_args(
        name="Chia Tea Monitoring Client",
        description=(
            "This tool collects data about chia and the machine"
            + "and sends them to the monitoring server."
        ),
    )

    # load config
    config = read_config(args.config)

    # create the watchdog
    watchdog = ChiaWatchdog(
        config.chia.logfile_filepath,
        config.chia.madmax_logfile,
    )

    # we disable auth during testing
    is_testing = config.development.testing

    cert = get_credentials_cert(is_testing, config) if not is_testing else ""

    # create client
    client = MonitoringClient(
        chia_dog=watchdog,
        config=config.monitoring.client,
        credentials_cert=cert,
        machine_name=config.machine.name,
    )

    # setup event loops
    loop = asyncio.get_event_loop()
    loop.create_task(run_watchdog(watchdog))
    loop.create_task(client.start_sending_updates())
    loop.run_forever()


if __name__ == "__main__":
    main()
